
#returns x such that y = trigamma(x)
#if X~gamma(shape=a,rate=b), then Var(ln(X)) = trigamma(a)
#thus this function is useful to find shape parameter which yields a desired variance of ln(X)
inv.trigamma = function(y,tol=1e-10) {
	x = 1
	error = y-trigamma(x);
	while(abs(error)>tol) {
		x = x + error/psigamma(x,2);#Newtons method
		error = y - trigamma(x);
	}
	list(x=x,accuracy=error)
}

observed.loglikelihood = function(ya,yb,xa,wa,wb,betas,beta0,sigma2,psi,diag.nu,tau2,mu.X,precision.X,G,g) {
	n.a=length(ya); n.b = length(yb); p = dim(s.xa)[2]
	if(missing(G)) {
		G = solve(betas%*%t(betas)/sigma2+diag.nu^2/tau2+precision.X)
	}
	if(missing(g)) {
		g = (yb-beta0)%*%t(betas)/sigma2+(wb-rep(psi,each=n.b))%*%diag.nu/tau2+matrix(mu.X,nrow=n.b,ncol=p,byrow=T)%*%precision.X
	}
	foo = xa-rep(mu.X,each=n.a)
	drop(-(n.a+n.b)*log(sigma2*(tau2^p))+(n.a+n.b)*determinant(precision.X)$modulus-
		sum((ya-beta0-xa%*%betas)^2)/sigma2-
		sum((wa-rep(psi,each=n.a)-xa%*%diag.nu)^2)/tau2-
		sum(diag(foo%*%precision.X%*%t(foo)))+
		n.b*determinant(G)$modulus-
		sum((yb-beta0)^2)/sigma2-sum((wb-rep(psi,each=n.b))*(wb-rep(psi,each=n.b)))/tau2-n.b*t(mu.X)%*%precision.X%*%mu.X+
		sum((g%*%G)*g))
}


em <- function(ya,yb,xa,wa,wb,
		location.factor,scale.factor,
		init.params,hyperparams,
		lambda.betas.update="none",lambda.precision.X.update="none",
		single.psi=T,single.nu=T,converge.tol=1e-3,s.xa) {
	
	n.a = length(ya); n.b = length(yb); p = ncol(xa);
	n = n.a + n.b;

	imp.betas = init.params$betas;
	imp.beta0 = init.params$beta0;
	scaled.imp.beta0 = imp.beta0 + sum(location.factor*imp.betas);
	imp.sigma2 = init.params$sigma2;
	imp.psi = init.params$psi;
	diag.imp.nu = diag(init.params$nu,p);
	imp.tau2 = init.params$tau2;
	imp.mu.X = init.params$mu.X;
	imp.precision.X = init.params$precision.X;
	
	imp.lambda.betas = hyperparams$lambda.betas;
	imp.lambda.precision.X = hyperparams$lambda.precision.X;
	precision.X.prior.df = hyperparams$precision.X.prior.df;
	if("lambda.betas.prior.shape"%in%names(hyperparams)) {lambda.betas.prior.shape = hyperparams$lambda.betas.prior.shape;
		} else {lambda.betas.prior.shape = 0;}
	if("lambda.betas.prior.rate"%in%names(hyperparams)) {lambda.betas.prior.rate = hyperparams$lambda.betas.prior.rate;
		} else {lambda.betas.prior.rate = 0;}
	if(lambda.betas.update=="ln") {
		log.lambda.grid = list(from=-7,to=7,length.out=50);
		lambda.seq = exp(log.lambda.seq<-do.call(seq,log.lambda.grid));
	}

	trace.imp.lambda.precision.X = sum(diag(imp.lambda.precision.X))

	old.params = list(max.betas=imp.betas,max.beta0=imp.beta0,max.sigma2=imp.sigma2,max.psi=imp.psi,max.nu=diag(diag.imp.nu),max.tau2=imp.tau2,max.mu.X=imp.mu.X,max.precision.X=imp.precision.X,max.lambda.betas=imp.lambda.betas,max.lambda.precision.X=imp.lambda.precision.X);

	crossprod.s.xa = crossprod(s.xa);
	crossprod.ya.s.xa = crossprod(ya,s.xa);
	diag.crossprod.xa.xa = colSums(xa^2);
	diag.crossprod.xa.wa = colSums(xa*wa);
	colSums.s.xa = colSums(s.xa);
	colSums.xa = colSums(xa);
	colSums.wa = colSums(wa);
	colSums.wb = colSums(wb);

	Ip = diag(1,p);

	j = 1;error=F;
	g = (yb-imp.beta0)%*%t(imp.betas)/imp.sigma2+(wb-rep(imp.psi,each=n.b))%*%diag.imp.nu/imp.tau2 + matrix(imp.mu.X,nrow=n.b,ncol=p,byrow=T)%*%imp.precision.X
	var.imp.xb = try(solve(imp.betas%*%t(imp.betas)/imp.sigma2+diag.imp.nu^2/imp.tau2+imp.precision.X),silent=T);
#obs.likes = observed.loglikelihood(ya,yb,xa,wa,wb,imp.betas,imp.beta0,imp.sigma2,imp.psi,diag.imp.nu,imp.tau2,imp.mu.X,imp.precision.X,var.imp.xb,g)

	small.number = .Machine$double.eps^.5

	while(j<1000) {

		j=j+1;

		#EStep
		if(class(var.imp.xb)=='try-error') {warning("SINGULAR MATRIX IN XB IMPUTATION");error=T;break;}	
		imp.xb = g%*%var.imp.xb
		imp.s.xb = scale(imp.xb,center=location.factor,scale=scale.factor);
		trace.var.imp.xb = sum(diag(var.imp.xb));

		#CMStep(i) -- betas
		#impute betas
		crossprod.imp.s.xb = crossprod(imp.s.xb)
		var.scaled.imp.betas = try(solve(crossprod.s.xa + crossprod.imp.s.xb + n.b*var.imp.xb/rep(scale.factor,each=p)/rep(scale.factor,times=p) + imp.lambda.betas*Ip),silent=T);
		if(class(var.scaled.imp.betas)=='try-error') {warning("SINGULAR MATRIX IN BETA STEP"); error=T; break;}
		scaled.imp.betas = drop((crossprod.ya.s.xa - scaled.imp.beta0*colSums.s.xa + (yb-scaled.imp.beta0)%*%imp.s.xb)%*%var.scaled.imp.betas);
		dotprod.scaled.imp.betas = sum(scaled.imp.betas^2)
		imp.betas = scaled.imp.betas/scale.factor;
#obs.likes = c(obs.likes,observed.loglikelihood(ya,yb,xa,wa,wb,imp.betas,imp.beta0,imp.sigma2,imp.psi,diag.imp.nu,imp.tau2,imp.mu.X,imp.precision.X))
		
		#CMStep(ii) -- beta0 
		reg.resid.a = ya-xa%*%imp.betas
		reg.resid.b = yb-imp.xb%*%imp.betas
		imp.beta0 = (sum(reg.resid.a)+sum(reg.resid.b))/n
		scaled.imp.beta0 = imp.beta0 + sum(imp.betas*location.factor);
#obs.likes = c(obs.likes,observed.loglikelihood(ya,yb,xa,wa,wb,imp.betas,imp.beta0,imp.sigma2,imp.psi,diag.imp.nu,imp.tau2,imp.mu.X,imp.precision.X))

		#CMStep(iii) -- sigma2
		reg.resid.a = reg.resid.a - imp.beta0
		reg.resid.b = reg.resid.b - imp.beta0
		imp.sigma2 = drop((sum(reg.resid.a^2)+sum(reg.resid.b^2)+n.b*t(imp.betas)%*%var.imp.xb%*%imp.betas+imp.lambda.betas*dotprod.scaled.imp.betas)/(n+p*(imp.lambda.betas>0)))
#obs.likes = c(obs.likes,observed.loglikelihood(ya,yb,xa,wa,wb,imp.betas,imp.beta0,imp.sigma2,imp.psi,diag.imp.nu,imp.tau2,imp.mu.X,imp.precision.X))


		#CMStep(iv) -- psi
		colSums.imp.xb = colSums(imp.xb);
		imp.psi = drop(colSums.wa+colSums.wb-diag.imp.nu%*%(colSums.xa+colSums.imp.xb))/n;
		if(single.psi) {
			imp.psi = rep(mean(imp.psi),p);
		} 
#obs.likes = c(obs.likes,observed.loglikelihood(ya,yb,xa,wa,wb,imp.betas,imp.beta0,imp.sigma2,imp.psi,diag.imp.nu,imp.tau2,imp.mu.X,imp.precision.X))

		#CMStep(v) -- nu
		numerator.imp.nu = diag.crossprod.xa.wa+colSums(imp.xb*wb)-colSums.xa*imp.psi-colSums.imp.xb*imp.psi;
		denominator.imp.nu = diag.crossprod.xa.xa+colSums(imp.xb*imp.xb)+n.b*diag(var.imp.xb)
		if(single.nu) {
			diag.imp.nu = diag(sum(numerator.imp.nu)/sum(denominator.imp.nu),p);
		} else {
			diag.imp.nu = diag(numerator.imp.nu/denominator.imp.nu);
		}
#obs.likes = c(obs.likes,observed.loglikelihood(ya,yb,xa,wa,wb,imp.betas,imp.beta0,imp.sigma2,imp.psi,diag.imp.nu,imp.tau2,imp.mu.X,imp.precision.X))

		#CMStep(vi) -- tau2
		imp.tau2 = (sum((wa-rep(imp.psi,each=n.a)-xa%*%diag.imp.nu)^2)+sum((wb-rep(imp.psi,each=n.b)-imp.xb%*%diag.imp.nu)^2)+n.b*sum(diag(diag.imp.nu^2)*diag(var.imp.xb)))/(p*n)
#obs.likes = c(obs.likes,observed.loglikelihood(ya,yb,xa,wa,wb,imp.betas,imp.beta0,imp.sigma2,imp.psi,diag.imp.nu,imp.tau2,imp.mu.X,imp.precision.X))

		#CMStep(vii) -- mu.X
		imp.mu.X = (colSums.xa + colSums.imp.xb)/n
#obs.likes = c(obs.likes,observed.loglikelihood(ya,yb,xa,wa,wb,imp.betas,imp.beta0,imp.sigma2,imp.psi,diag.imp.nu,imp.tau2,imp.mu.X,imp.precision.X))
				
		#CMStep(viii) -- precision.X
		imp.precision.X  = try((n + (precision.X.prior.df-p-1)*(trace.imp.lambda.precision.X!=0))*solve(crossprod(xa-rep(imp.mu.X,each=n.a)) + crossprod(imp.xb-rep(imp.mu.X,each=n.b)) + n.b*var.imp.xb + imp.lambda.precision.X),silent=T)
		if(class(imp.precision.X)=='try-error') {warning("SINGULAR MATRIX IN PRECISION.X STEP"); error=T; break;}
#obs.likes = c(obs.likes,observed.loglikelihood(ya,yb,xa,wa,wb,imp.betas,imp.beta0,imp.sigma2,imp.psi,diag.imp.nu,imp.tau2,imp.mu.X,imp.precision.X))

		#CMStep(ix) -- lambda
		if(lambda.betas.update=="ga") {
			imp.lambda.betas = (p+2*lambda.betas.prior.shape-2)/(dotprod.scaled.imp.betas/imp.sigma2+2*lambda.betas.prior.rate)
		} else if(lambda.betas.update=="ln") {
			imp.lambda.betas = lambda.seq[which.max(p*log.lambda.seq-lambda.seq*dotprod.scaled.imp.betas/imp.sigma2 - 2*log.lambda.seq - (log.lambda.seq +log(lambda.betas.prior.rate))^2/lambda.betas.prior.shape)]
			log.imp.lambda.betas = log(imp.lambda.betas);
			log.lambda.grid$from = log.imp.lambda.betas-(j<=20)-1/(5*j);
			log.lambda.grid$to = log.imp.lambda.betas+(j<=20)+1/(5*j);
			lambda.seq = exp(log.lambda.seq<-do.call(seq,log.lambda.grid));
		} else if(lambda.betas.update=="ig") {
			foo = dotprod.scaled.imp.betas/imp.sigma2
			imp.lambda.betas = (p-2*lambda.betas.prior.shape-2+sqrt((p-2*lambda.betas.prior.shape-2)^2+8*foo/lambda.betas.prior.rate))/(2*foo)
		}

		#CMStep(x) -- psi 
		#if(psi.update=="eb") {
		#	imp.psi = (precision.X.prior.df-p-1) * solve(imp.precision.X);
		#}

		
		g = (yb-imp.beta0)%*%t(imp.betas)/imp.sigma2+(wb-rep(imp.psi,each=n.b))%*%diag.imp.nu/imp.tau2 + matrix(imp.mu.X,nrow=n.b,ncol=p,byrow=T)%*%imp.precision.X
		var.imp.xb = try(solve(imp.betas%*%t(imp.betas)/imp.sigma2+diag.imp.nu^2/imp.tau2+imp.precision.X),silent=T);
		new.params = list(max.betas=imp.betas,max.beta0=imp.beta0,max.sigma2=imp.sigma2,max.psi=imp.psi,max.nu=diag(diag.imp.nu),max.tau2=imp.tau2,max.mu.X=imp.mu.X,max.precision.X=imp.precision.X,max.lambda.betas=imp.lambda.betas,max.lambda.precision.X=imp.lambda.precision.X);

		if(max(log(abs((unlist(new.params)+small.number)/(unlist(old.params)+small.number))))<converge.tol) {break;}
		old.params = new.params;
#print(imp.lambda.betas);

	}
	new.params$j = j
	if(error) { NA; } else { return(new.params); }
}	

rwish.phil <- function (v, inv.chol.invscale) 
{
	p=nrow(inv.chol.invscale)
	if (!is.matrix(inv.chol.invscale)) {
		stop(message = "inv.chol.invscale not a matrix in rwish.phil().\n")
		}
	if (p != ncol(inv.chol.invscale)) {
      	stop(message = "inv.chol.invscale not square in rwish.phil().\n")
	}
	if (v < p) {
      	stop(message = "v is less than the dimension of inv.chol.invscale in rwish.phil().\n")
	}
    	Z <- matrix(0, p, p)
	diag(Z) <- sqrt(rchisq(p, v:(v - p + 1)))
	if (p > 1) {
      	pseq <- 1:(p - 1)
		Z[upper.tri(Z)] <- rnorm(p * (p - 1)/2)
	}
	return(crossprod(Z %*% inv.chol.invscale))
}	

ldwish.dfpart = function(v,C,p) {
	(v/2)*C - sum(lgamma((v + 1 - 1:p)/2))
}

ldwish.dfpart = Vectorize(ldwish.dfpart,"v")


da.gibbs <- function(ya,yb,xa,wa,wb,
				location.factor,scale.factor,
				init.params,hyperparams,mcmc.burnin,mcmc.niter,
				lambda.betas.update="none",lambda.precision.X.update="none",
				single.psi=T,single.nu=T,constrain.lambda.precision.X="none",s.xa) {
	
	n.a = length(ya); n.b = length(yb); p = dim(xa)[2]
	n = n.a + n.b;

	imp.betas = init.params$betas;
	imp.beta0 = init.params$beta0;
	scaled.imp.beta0 = imp.beta0 + sum(location.factor*imp.betas);
	imp.sigma2 = init.params$sigma2;
	imp.psi = init.params$psi;
	diag.imp.nu = diag(init.params$nu,p);
	imp.tau2 = init.params$tau2;
	imp.mu.X = init.params$mu.X;
	imp.precision.X = init.params$precision.X;

	imp.lambda.betas = hyperparams$lambda.betas;
	imp.lambda.precision.X = hyperparams$lambda.precision.X;
	precision.X.prior.df = hyperparams$precision.X.prior.df;
	if("lambda.betas.prior.shape"%in%names(hyperparams)) {lambda.betas.prior.shape = hyperparams$lambda.betas.prior.shape;
		} else {lambda.betas.prior.shape = 0;}
	if("lambda.betas.prior.rate"%in%names(hyperparams)) {lambda.betas.prior.rate = hyperparams$lambda.betas.prior.rate;
		} else {lambda.betas.prior.rate = 0;}


	store.betas = matrix(0,nrow=mcmc.niter,ncol=p);
	store.beta0 = numeric(mcmc.niter);
	store.sigma2 = numeric(mcmc.niter);
	store.tau2 = numeric(mcmc.niter);
	store.lambda.betas = numeric(mcmc.niter);
	store.trace.lambda.precision.X = numeric(mcmc.niter);
	store.precision.X.prior.df = numeric(mcmc.niter);
	return.expxxT = matrix(0,nrow=p,ncol=p);
	return.expxxTbetas = matrix(0,nrow=p,ncol=1);

	if(lambda.betas.update=="eb") {
		n.ebiter.betas=10;
		m1=0;
		store.betatopbeta.over.sigma2 = 0;
	}

	if(grepl("eb",lambda.precision.X.update)) {
		n.ebiter.precision.X=ceiling(p/4);
		m2=0;
		store.precision.X = 0
		store.log.det.precision.X = 0;
		prior.df.grid = seq(max(p,precision.X.prior.df-5*precision.X.prior.df+1),precision.X.prior.df+5*precision.X.prior.df,by=round(precision.X.prior.df/10));#seq(max(p,precision.X.prior.df-5*p+1),precision.X.prior.df+5*p,by=1);#
	}

	trace.imp.lambda.precision.X = sum(diag(imp.lambda.precision.X))

	j = 0;error = F;
	
	Ip = diag(1,p)

	crossprod.s.xa = crossprod(s.xa);
	crossprod.ya.s.xa = crossprod(ya,s.xa);
	colSums.s.xa = colSums(s.xa);

	diag.crossprod.xa.xa = colSums(xa^2);
	diag.crossprod.xa.wa = colSums(xa*wa);
	colSums.xa = colSums(xa);
	colSums.wa = colSums(wa);
	colSums.wb = colSums(wb);

	while(j<2*n.ebiter.precision.X-1){#mcmc.burnin+mcmc.niter) {#

		#cat(j,'\n')
		j=j+1;
		#impute xb
		precision.imp.xb = imp.betas%*%t(imp.betas)/imp.sigma2+diag.imp.nu^2/imp.tau2+imp.precision.X
		chol.precision.imp.xb = try(t(chol(precision.imp.xb)),silent=T)
		if(class(chol.precision.imp.xb)=='try-error') {warning("SINGULAR MATRIX IN XB IMPUTATION");error=T;break;}	
		inv.chol.precision.imp.xb = solve(chol.precision.imp.xb)
		mean.imp.xb = ((yb-imp.beta0)%*%t(imp.betas)/imp.sigma2+(wb-rep(imp.psi,each=n.b))%*%diag.imp.nu/imp.tau2 + matrix(imp.mu.X,nrow=n.b,ncol=p,byrow=T)%*%imp.precision.X)%*%t(inv.chol.precision.imp.xb)
		imp.xb = (mean.imp.xb + matrix(rnorm(n.b*p),nrow=n.b))%*%inv.chol.precision.imp.xb
		imp.s.xb = scale(imp.xb,center=location.factor,scale=scale.factor);
		
		#impute betas
		crossprod.imp.s.xb = crossprod(imp.s.xb)
		precision.scaled.imp.betas = crossprod.s.xa + crossprod.imp.s.xb + imp.lambda.betas*Ip;
		chol.precision.scaled.imp.betas = try(t(chol(precision.scaled.imp.betas)),silent=T)
		if(class(chol.precision.scaled.imp.betas)=='try-error') {warning("SINGULAR MATRIX IN BETA IMPUTATION");error=T;break;}
		inv.chol.precision.scaled.imp.betas = solve(chol.precision.scaled.imp.betas);
		mean.scaled.imp.betas = (crossprod.ya.s.xa - scaled.imp.beta0*colSums.s.xa + (yb-scaled.imp.beta0)%*%imp.s.xb)%*%t(inv.chol.precision.scaled.imp.betas)
		scaled.imp.betas = drop((mean.scaled.imp.betas + sqrt(imp.sigma2)*rnorm(p))%*%inv.chol.precision.scaled.imp.betas);
		dotprod.scaled.imp.betas = sum(scaled.imp.betas^2)

		imp.betas = scaled.imp.betas/scale.factor;

		#beta0 
		reg.resid.a = ya-xa%*%imp.betas
		reg.resid.b = yb-imp.xb%*%imp.betas
		imp.beta0 = (sum(reg.resid.a)+sum(reg.resid.b))/n + sqrt(imp.sigma2/n)*rnorm(1)
		scaled.imp.beta0 = imp.beta0 + sum(imp.betas*location.factor);

		#sigma2
		reg.resid.a = reg.resid.a - imp.beta0
		reg.resid.b = reg.resid.b - imp.beta0
		imp.sigma2 = 1/rgamma(1,.5*(p*(imp.lambda.betas>0)+n),.5*(imp.lambda.betas*dotprod.scaled.imp.betas+sum(reg.resid.a^2)+sum(reg.resid.b^2)))

		#psi
		colSums.imp.xb = colSums(imp.xb);
		mean.imp.psi = drop(colSums.wa+colSums.wb-diag.imp.nu%*%(colSums.xa+colSums.imp.xb))/n;
		if(single.psi) {
			imp.psi = rep(mean(mean.imp.psi) + sqrt(imp.tau2/n/p)*rnorm(1),p);
		} else {
			imp.psi = mean.imp.psi + sqrt(imp.tau2/n)*rnorm(p);
		}

		#nu
		numerator.mean.imp.nu = diag.crossprod.xa.wa+colSums(imp.xb*wb)-colSums.xa*imp.psi-colSums.imp.xb*imp.psi;
		denominator.mean.imp.nu = diag.crossprod.xa.xa+colSums(imp.xb*imp.xb);
		if(single.nu) {
			denominator.mean.imp.nu = sum(denominator.mean.imp.nu)
			diag.imp.nu = diag(sum(numerator.mean.imp.nu)/denominator.mean.imp.nu + sqrt(imp.tau2/denominator.mean.imp.nu)*rnorm(1),p);
		} else {
			diag.imp.nu = diag(numerator.mean.imp.nu/denominator.mean.imp.nu + sqrt(imp.tau2/denominator.mean.imp.nu)*rnorm(p),p);
		}

		#tau2
		imp.tau2 = 1/rgamma(1,.5*n*p,.5*(sum((wa-rep(imp.psi,each=n.a)-xa%*%diag.imp.nu)^2)+sum((wb-rep(imp.psi,each=n.b)-imp.xb%*%diag.imp.nu)^2)))

		#precision.X
		invscale.imp.precision.X = crossprod(xa-rep(imp.mu.X,each=n.a)) + crossprod(imp.xb-rep(imp.mu.X,each=n.b)) + imp.lambda.precision.X
		chol.invscale.imp.precision.X = try(t(chol(invscale.imp.precision.X)),silent=T)
		if(class(chol.invscale.imp.precision.X)=='try-error') {warning("SINGULAR MATRIX IN PRECISION.X IMPUTATION");error=T;break;}	
		inv.chol.invscale.imp.precision.X = solve(chol.invscale.imp.precision.X)
		imp.precision.X = rwish.phil(n+precision.X.prior.df*(trace.imp.lambda.precision.X!=0),inv.chol.invscale.imp.precision.X)

		#mu.X
		chol.imp.precision.X = try(t(chol(imp.precision.X)),silent=T)
		if(class(chol.imp.precision.X)=='try-error') {warning("SINGULAR MATRIX IN MEAN.X STEP");error=T;break;}	
		inv.chol.imp.precision.X = solve(chol.imp.precision.X)
		imp.mu.X = drop(rnorm(p)%*%inv.chol.imp.precision.X/sqrt(n)+colSums.xa/n+colSums.imp.xb/n)

		#lambda.betas
		if(lambda.betas.update=="bayes") {
			imp.lambda.betas = rgamma(1,p/2+lambda.betas.prior.shape,dotprod.scaled.imp.betas/imp.sigma2/2+lambda.betas.prior.rate)
		} else if(lambda.betas.update=="eb") {
			m1=m1+1;
			store.betatopbeta.over.sigma2 = store.betatopbeta.over.sigma2 + dotprod.scaled.imp.betas/imp.sigma2
			if(m1==n.ebiter.betas) {
				imp.lambda.betas = (p*n.ebiter.betas)/(store.betatopbeta.over.sigma2)
				if(is.na(imp.lambda.betas)) { error=T; break; }
				store.betatopbeta.over.sigma2 = 0;
				m1 = 0;
				if(j>mcmc.burnin) {n.ebiter.betas=50;}
			}
		}

		#lambda.precision.X 
		if(lambda.precision.X.update=="eb.invscale") {
			m2=m2+1;
			if(constrain.lambda.precision.X=="diag") {store.precision.X = store.precision.X + diag(diag(imp.precision.X));
				} else if(constrain.lambda.precision.X=="scalar") {store.precision.X = mean(diag(imp.precision.X))*Ip;
			}
			if(m2==n.ebiter.precision.X) {
				imp.lambda.precision.X = (precision.X.prior.df*n.ebiter.precision.X)*solve(store.precision.X);
				trace.imp.lambda.precision.X = sum(diag(imp.lambda.precision.X));
				store.precision.X = 0;
				if(is.na(trace.imp.lambda.precision.X)) { error=T; break; }
				m2=0;
				#if(j>mcmc.burnin) {n.ebiter.precision.X=3*p;}
				if(j>1000) {n.ebiter.precision.X=3*p;}
			}
		} else if(lambda.precision.X.update=="eb.dof") {
			m2=m2+1;
			store.log.det.precision.X = store.log.det.precision.X + as.numeric(determinant(imp.precision.X,log=T)$mod)
			if(m2==n.ebiter.precision.X) {
				precision.X.prior.df = prior.df.grid[which.max(ldwish.dfpart(prior.df.grid,p*log(2)+as.numeric(determinant(imp.lambda.precision.X,log=T)$mod)+store.log.det.precision.X/n.ebiter.precision.X,p))]
				prior.df.grid = seq(max(p,precision.X.prior.df-5*precision.X.prior.df+1),precision.X.prior.df+5*precision.X.prior.df,by=round(precision.X.prior.df/10));#seq(max(p,precision.X.prior.df-5*p+1),precision.X.prior.df+5*p,by=1);#
				store.log.det.precision.X = 0;
				m2=0;
				#if(j>mcmc.burnin) {n.ebiter.precision.X=3*p;}
				if(j>1000) {n.ebiter.precision.X=3*p;}
			}
		}
		if(j%%100==0) {cat(j,'\n')}

		if(j>mcmc.burnin) {
			store.beta0[j-mcmc.burnin] = imp.beta0;
			store.betas[j-mcmc.burnin,] = imp.betas;
			store.sigma2[j-mcmc.burnin] = imp.sigma2;
			store.tau2[j-mcmc.burnin] = imp.tau2;
			store.lambda.betas[j-mcmc.burnin] = imp.lambda.betas;
			store.trace.lambda.precision.X[j-mcmc.burnin] = trace.imp.lambda.precision.X;
			store.precision.X.prior.df[j-mcmc.burnin] = precision.X.prior.df;
			foo = imp.mu.X%*%t(imp.mu.X)+t(inv.chol.imp.precision.X)%*%inv.chol.imp.precision.X;
			return.expxxT = return.expxxT + foo;
			return.expxxTbetas = return.expxxTbetas + foo%*%imp.betas;
			rm(foo);
			if(j==mcmc.burnin+1) {colnames(store.betas)=seq(1,p,by=1);}
		}

	}
	return.betas = colMeans(store.betas);
	return.beta0 = mean(store.beta0);
	
	if(error) {	
		rep(NA,p+1) 
	} else {
		list(	mean.betas.qloss=return.betas,
			mean.betas.ploss=drop(solve(return.expxxT)%*%return.expxxTbetas),
			mean.beta0 = return.beta0,
			store.betas = store.betas,
			store.beta0 = store.beta0,
			store.sigma2 = store.sigma2,
			store.tau2 = store.tau2,
			store.lambda.betas = store.lambda.betas,
			store.trace.lambda.precision.X=store.trace.lambda.precision.X,
			store.precision.X.prior.df=store.precision.X.prior.df) 
	}
}



da.gibbs.C <- function(ya,yb,xa,wa,wb,
		location.factor,scale.factor,
		init.params,hyperparams,mcmc.burnin,mcmc.niter,
		lambda.betas.update="none",lambda.precision.X.update="none",
		single.psi,single.nu,constrain.lambda.precision.X,s.xa) {
	
	p = ncol(xa);
	n.a = length(ya);
	n.b = length(yb);

	betas = init.params$betas;
	if(length(betas)!=p) {stop("initial betas must be length p");}
	beta0 = init.params$beta0;
	sigma2 = init.params$sigma2;
	psi = init.params$psi;
	nu = init.params$nu;
	tau2 = init.params$tau2;
	mu.X = init.params$mu.X;
	precision.X = init.params$precision.X;

	lambda.betas = hyperparams$lambda.betas;
	lambda.precision.X = hyperparams$lambda.precision.X;
	precision.X.prior.df = hyperparams$precision.X.prior.df;
	if("lambda.betas.prior.shape"%in%names(hyperparams)) {lambda.betas.prior.shape = hyperparams$lambda.betas.prior.shape;
		} else {lambda.betas.prior.shape = 0;}
	if("lambda.betas.prior.rate"%in%names(hyperparams)) {lambda.betas.prior.rate = hyperparams$lambda.betas.prior.rate;
		} else {lambda.betas.prior.rate = 0;}

	if(lambda.betas.update=="eb") {lambda.betas.update=1;
	} else if(lambda.betas.update=="bayes") {lambda.betas.update=2;
	} else {lambda.betas.update=0;}

	if(lambda.precision.X.update=="eb.invscale") {lambda.precision.X.update=1;
	} else if(lambda.precision.X.update=="eb.dof") {lambda.precision.X.update=2;	
	} else {lambda.precision.X.update=0;}

	if(constrain.lambda.precision.X=="diag") {constrain.lambda.precision.X=1;
	} else if(constrain.lambda.precision.X=="scalar") {constrain.lambda.precision.X=2;
	} else {stop("constrain.lambda.precision.X must be 'diag' or 'scalar'");}

	if(single.psi&single.nu) {
		psi=psi[1];
		nu=nu[1];
	}

	tmp  = .C("R_dagibbs",
			as.integer(p),as.integer(n.a),as.integer(n.b),
			as.double(ya),as.double(yb),as.double(t(xa)),as.double(t(wa)),as.double(t(wb)),
			as.double(location.factor),as.double(scale.factor),
			as.double(t(wb)),as.double(betas),
			as.double(beta0),as.double(sigma2),
			as.double(psi),as.double(nu),as.double(tau2),
			as.double(mu.X),as.double(precision.X),
			as.double(lambda.betas),as.double(lambda.precision.X),
			as.integer(precision.X.prior.df),
			as.integer(mcmc.burnin),as.integer(mcmc.niter),
			as.integer(lambda.betas.update),as.integer(lambda.precision.X.update),
			as.double(lambda.betas.prior.shape),as.double(lambda.betas.prior.rate),
			as.double(rep(0,p)),as.double(rep(0,p)),
			as.double(0),
			as.double(rep(0,mcmc.niter*p)),
			as.double(rep(0,mcmc.niter)),
			as.double(rep(0,mcmc.niter)),
			as.double(rep(0,mcmc.niter)),
			as.double(rep(0,mcmc.niter)),
			as.double(rep(0,mcmc.niter*p)),
			as.double(rep(0,mcmc.niter)),
			as.double(rep(0,mcmc.niter)),
			as.integer(0),as.integer(sample(2^30.9,1)),
			as.integer(single.psi),as.integer(single.nu),
			as.integer(constrain.lambda.precision.X));
	
	if(tmp[[40]]==0) {
		list(mean.betas.qloss=tmp[[29]],
			mean.betas.ploss=tmp[[30]],
			mean.beta0=tmp[[31]],
			store.betas=matrix(tmp[[32]],nrow=mcmc.niter,byrow=T),
			store.beta0=tmp[[33]],
			store.sigma2=tmp[[34]],
			store.tau2=tmp[[35]],
			store.lambda.betas=tmp[[36]],
			store.lambda.precision.X=matrix(tmp[[37]],nrow=mcmc.niter,byrow=T),
			store.trace.lambda.precision.X=tmp[[38]],
			store.precision.X.prior.df=tmp[[39]]);
	} else if(tmp[[40]]==1) {
		warning("SINGULAR MATRIX IN XB IMPUTATION (C)");
		list(betas.qloss=rep(NA,p),betas.ploss=rep(NA,p));
	} else if(tmp[[40]]==2) {
		warning("SINGULAR MATRIX IN BETA IMPUTATION (C)");
		list(betas.qloss=rep(NA,p),betas.ploss=rep(NA,p));
	} else if(tmp[[40]]==3) {
		warning("SINGULAR MATRIX IN PRECISION.X IMPUTATION (C)");
		list(betas.qloss=rep(NA,p),betas.ploss=rep(NA,p));
	} else if(tmp[[40]]==4) {
		warning("SINGULAR MATRIX IN MU.X IMPUTATION (C)");
		list(betas.qloss=rep(NA,p),betas.ploss=rep(NA,p));
	} else if(tmp[[40]]==5) {
		warning("ZERO VALUE IN LAMBDA.BETAS EMPIRICAL BAYES IMPUTATION (C)");
		list(betas.qloss=rep(NA,p),betas.ploss=rep(NA,p));
	} else if(tmp[[40]]==6) {
		warning("SINGULAR MATRIX IN LAMBDA.PRECISION.X EMPIRICAL BAYES IMPUTATION (C)");
		list(betas.qloss=rep(NA,p),betas.ploss=rep(NA,p));
	} else if(tmp[[40]]==7) {
		warning("SINGULAR MATRIX IN XB IMPUTATION (C)");
		list(betas.qloss=rep(NA,p),betas.ploss=rep(NA,p));
	}
	
}

if(0) {




score.B = function(betas,yb,imp.s.xb,wb,sigma2,nu,tau2,precision.X) {
	n.b = length(yb); p = dim(wb)[2]
	V = solve(diag(1,p)+tau2*precision.X/nu^2)
	kappa = drop(tau2/nu^2*t(betas)%*%V%*%betas)
	betas.top.V = drop(t(betas)%*%V)
	return1 = -n.b*(tau2/nu^2)/(sigma2+kappa)*betas.top.V
	return2 = (sigma2)^(-2)*(tau2/nu^2)*sum(yb^2)*betas.top.V + (sigma2)^(-1)*colSums(yb*imp.s.xb)
	return3 = (tau2/nu^2)*(sigma2+kappa)^(-2)*sum((yb*kappa/sigma2+imp.s.xb%*%betas)^2)*betas.top.V
	return4 = -(sigma2+kappa)^(-1)*sum((yb*kappa/sigma2+imp.s.xb%*%betas)*yb)*2*(sigma2)^(-1)*(tau2/nu^2)*betas.top.V-
			(sigma2+kappa)^(-1)*colSums(drop(yb*kappa/sigma2+imp.s.xb%*%betas)*imp.s.xb)
	return(return1+return2+return3+return4)
	
}

score.sq.B = function(betas,yb,imp.s.xb,wb,sigma2,nu,tau2,precision.X) {
	foo = score.B(betas,yb,imp.s.xb,wb,sigma2,nu,tau2,precision.X)
	drop(sum(foo^2))
}



min.lambda = function(p,yb,imp.s.xb,wb,sigma2,nu,tau2,precision.X) {
	foo = sqrt(sigma2)*matrix(rnorm(p*p*10),ncol=p)
	lambda.seq = seq(1e-5,1001,by=10)
	denoms = numeric(length(lambda.seq))
	for(i in 1:length(lambda.seq)) {
		rand.betas = foo/sqrt(lambda.seq[i])
		denoms[i] = mean(apply(rand.betas,1,score.sq.B,yb=yb,imp.s.xb=imp.s.xb,wb=wb,sigma2=sigma2,nu=nu,tau2=tau2,precision.X=precision.X))+p*lambda.seq[i]/sigma2
	}
	denoms
	#imax = which.max(denoms)
	#if(imax==1|imax==length(lambda.seq)) {
	#	return(lambda.seq[imax]) 
	#} else {
	#	lambda.seq = sort(runif(10,lambda.seq[imax-1],lambda.seq[imax+1]))
	#	denoms = numeric(length(lambda.seq))
	#	for(i in 1:length(lambda.seq)) {
	#		rand.betas = foo/sqrt(lambda.seq[i])
	#		denoms[i] = mean(apply(rand.betas,1,score.sq.B,yb=yb,imp.s.xb=imp.s.xb,wb=wb,sigma2=sigma2,nu=nu,tau2=tau2,precision.X=solve(s.varcov.X)))+p*lambda.seq[i]/sigma2
	#	}
	#}
}
}

