/* To do:
 *
 *
 *
 *  
 */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <Rconfig.h>
#include <R.h>
#include <Rmath.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_statistics_double.h>
#include <gsl/gsl_sf_gamma.h>

//returns the log of the wishart density (only the expressions containing the degrees of freedom parameter). 
double gsl_ran_wishart_lpdf(const int df, const double C, const int dim) {
	int i;
	double multivariate_lngamma = 0.;
	for(i = 1; i<=dim; i++) {
		multivariate_lngamma += gsl_sf_lngamma((double)(df+1.-i)/2.);
	}
	return (df/2.)*C - multivariate_lngamma;
}

//for a grid of values of the degrees of freedom, df_grid, finds the values which returns the largest wishart density
int best_df(int *df_grid, const int length_df_grid, const double C, const int dim) {
	int i;
	int curr_which_max = 0;
	double curr_try = -HUGE_VAL;
	double curr_max = -HUGE_VAL;
	for(i = 0;i<length_df_grid;i++) {
		if(df_grid[i]<dim) {
			curr_try = -HUGE_VAL; 
		} else {
			curr_try = gsl_ran_wishart_lpdf(df_grid[i],C,dim);
		}
		if(curr_try>curr_max) {
			curr_which_max = i;			
			curr_max = curr_try;
		}
	}
	return df_grid[curr_which_max];
}

//Add the diagonal elements of square matrix MAT to square matrix RESULT
void gsl_matrix_add_diagonals(const int ndim, gsl_matrix * RESULT, const gsl_matrix * MAT) {
	int i;
	for(i=0; i<ndim; i++) {
		gsl_matrix_set(RESULT,i,i,gsl_matrix_get(RESULT,i,i)+gsl_matrix_get(MAT,i,i));	
	}
}

//Add a constant double x to square matrix RESULT
void gsl_matrix_add_constant_diagonal_only(const int ndim, gsl_matrix * RESULT, const double x) {
	int i;
	for(i=0; i<ndim; i++) {
		gsl_matrix_set(RESULT,i,i,gsl_matrix_get(RESULT,i,i)+x);	
	}
}

//
void matrix_crossprod(const int nrows, const int ncols, const gsl_matrix * MAT, gsl_matrix * RESULT ) {
	int i,j,k;
	gsl_matrix_set_zero (RESULT);
	for(k=0; k<nrows; k++) {
		for(i=0; i<ncols; i++) for(j=0;j<ncols; j++) {
			gsl_matrix_set(RESULT,i,j,gsl_matrix_get(RESULT,i,j)+gsl_matrix_get(MAT,k,i)*gsl_matrix_get(MAT,k,j));	
		}
	}
}


//Puts the column sums of an (nrows)x(ncols) gsl_matrix MAT into a (ncols)x(1) gsl_matrix RESULT
void column_sums(const int nrows, const int ncols, const gsl_matrix * MAT, gsl_matrix * RESULT) {
	int i,j;
	gsl_matrix_set_zero (RESULT);
	for(i=0; i<nrows; i++) {
		for(j=0; j<ncols; j++) {
			gsl_matrix_set(RESULT,j,0,gsl_matrix_get(RESULT,j,0)+gsl_matrix_get(MAT,i,j));	
		}
	}
}

//returns the sum of the diagonal elements of a gsl_matrix MAT
double matrix_trace(const int dim, const gsl_matrix * MAT) {
	int i;
	double tr = 0.;
	for(i=0; i<dim; i++) {tr += gsl_matrix_get(MAT,i,i);}
	return tr;
}

//returns the element-wise sum of an (nrows)x(ncols) gsl_matrix MAT, with each element raised to the exponent power
double element_wise_sum(const int nrows, const int ncols, const gsl_matrix * MAT, const double exponent ) {
	int i,j; 
	double return_val = 0.;
	for(i=0; i<nrows; i++) {
		for(j=0; j<ncols; j++) {
			return_val += pow(gsl_matrix_get(MAT,i,j),exponent);
		}
	}
	return return_val;
}


//Puts a random wishart matrix with df degrees of freedom and scale matrix (INV_CHOL_INVSCALE)^T%*%(INV_CHOL_INVSCALE)
//[that is, the gsl_matrix INV_CHOL_INVSCALE is assumed to be the inverse of the lower triangular cholesky matrix of the 
//INVERSE scale matrix] into the (dim)x(dim) matrix gsl_matrix RESULT
double wishart(const int df, gsl_matrix * INV_CHOL_INVSCALE, const int dim, gsl_rng * ran_num_gen, gsl_matrix * RESULT) {
	int i,j;
	gsl_matrix_set_zero (RESULT);
	gsl_matrix * PLACEHOLDER = gsl_matrix_calloc(dim,dim);
	for(i=0;i<dim;i++) {
		gsl_matrix_set (PLACEHOLDER, i, i,sqrt(gsl_ran_chisq(ran_num_gen,df-i)));
		for(j=0;j<i;j++) {
			gsl_matrix_set (PLACEHOLDER, i, j, gsl_ran_ugaussian(ran_num_gen));
		}
	}
	double lndet_RESULT = 2.*gsl_linalg_LU_lndet(PLACEHOLDER)+2.*gsl_linalg_LU_lndet(INV_CHOL_INVSCALE);

	gsl_blas_dtrmm (CblasLeft, CblasLower,CblasTrans,CblasNonUnit,1.,INV_CHOL_INVSCALE,PLACEHOLDER);
	gsl_blas_dgemm(CblasNoTrans,CblasTrans,1.,PLACEHOLDER,PLACEHOLDER,0.,RESULT);

	gsl_matrix_free (PLACEHOLDER);

	return lndet_RESULT;
}


void dagibbs(int p, int na, int nb,
		const double *ya,const double *yb,const double *xa,const double *wa,const double *wb,
		const double *x_location_factor, const double *x_scale_factor,
		double *imp_xb, double *imp_betas,
		double *imp_beta0, double *imp_sigmasq, 
		double *imp_psi, double *imp_nu, double *imp_tausq, 
		double *imp_mean_x, double *imp_precision_x,
		double *imp_lambda_betas, double *imp_lambda_precision_x,
		int *precision_x_prior_df,
		const int mcmc_burnin, const int mcmc_niter, 
		const int lambda_betas_update, const int lambda_precision_x_update,
		const double lambda_betas_prior_shape, const double lambda_betas_prior_rate,
		double *return_betas_qloss, double *return_betas_ploss,
		double *return_beta0,
		double *store_betas,
		double *store_beta0,
		double *store_sigmasq,
		double *store_tausq,
		double *store_lambda_betas,
		double *store_lambda_precision_x,
		double *store_trace_lambda_precision_x,
		double *store_precision_x_prior_df,
		int *error_code, const int random_seed,
		const int constrain_lambda_precision_x)
{
	int i, j, k;//index integers

	gsl_set_error_handler_off();

	//Initialize seed
	gsl_rng * ran_num_gen = gsl_rng_alloc(gsl_rng_mt19937);
	gsl_rng_set(ran_num_gen,random_seed);

	//Translate data to GSL matrices
	gsl_matrix * YA = gsl_matrix_calloc(na,1);
	for(i=0; i<na; i++) {gsl_matrix_set(YA, i, 0, ya[i]);}
	gsl_matrix * YB = gsl_matrix_calloc(nb,1);
	for(i=0; i<nb; i++) {gsl_matrix_set(YB, i, 0, yb[i]);}
	gsl_matrix * XA = gsl_matrix_calloc(na,p);
	for(i=0; i<na; i++) for(j=0; j<p; j++) {gsl_matrix_set(XA, i, j, xa[i*p+j]);}
	gsl_matrix * WA = gsl_matrix_calloc(na,p);
	for(i=0; i<na; i++) for(j=0; j<p; j++) {gsl_matrix_set(WA, i, j, wa[i*p+j]);}
	gsl_matrix * WB = gsl_matrix_calloc(nb,p);
	for(i=0; i<nb; i++) for(j=0; j<p; j++) {gsl_matrix_set(WB, i, j, wb[i*p+j]);}
	gsl_matrix * IMP_XB = gsl_matrix_calloc(nb,p);
	for(i=0; i<nb; i++) for(j=0; j<p; j++) {gsl_matrix_set(IMP_XB, i, j, imp_xb[i*p+j]);}
	//

	//Scaled versions
	gsl_matrix * X_LOCATION_FACTOR = gsl_matrix_calloc(p,1);
	for(i=0; i<p; i++) {gsl_matrix_set(X_LOCATION_FACTOR, i, 0, x_location_factor[i]);}
	gsl_matrix * X_INV_SCALE_FACTOR = gsl_matrix_calloc(p,p);
	for(i=0; i<p; i++) {gsl_matrix_set(X_INV_SCALE_FACTOR, i, i, 1./x_scale_factor[i]);}
	gsl_matrix * ONES_LENGTH_NA = gsl_matrix_calloc(na,1);//vector of 1's of length na
	gsl_matrix_set_all(ONES_LENGTH_NA,1.);
	gsl_matrix * ONES_LENGTH_NB = gsl_matrix_calloc(nb,1);//vector of 1's of length nb
	gsl_matrix_set_all(ONES_LENGTH_NB,1.);
	gsl_matrix * S_XA = gsl_matrix_calloc(na,p);
	gsl_matrix_memcpy(S_XA,XA);
	gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,ONES_LENGTH_NA,X_LOCATION_FACTOR,1.,S_XA);
	gsl_blas_dtrmm (CblasRight, CblasLower,CblasNoTrans,CblasNonUnit,1.,X_INV_SCALE_FACTOR,S_XA);
	gsl_matrix * IMP_S_XB = gsl_matrix_calloc(nb,p);
	//

	//Utility values that are used in each iteration
	//xb
	gsl_matrix * MEAN_IMP_XB = gsl_matrix_calloc(nb,p);//mean matrix of imputations
	gsl_matrix * CHOL_PRECISION_IMP_XB = gsl_matrix_calloc(p,p);//Lower triangular L such that Variance(imp_xb[i,]) = (L^-1)^T %*% (L^-1)
	gsl_matrix * INV_CHOL_PRECISION_IMP_XB = gsl_matrix_calloc(p,p);//L^(-1)
	gsl_matrix * RANDOM_VALS_IMP_XB = gsl_matrix_calloc(nb,p);//stochastic component
	gsl_matrix * CROSSPROD_IMP_MEAN_X_IMP_PRECISION_X = gsl_matrix_calloc(1,p);//t(mean_x)%*%precision_x

	//betas
	gsl_matrix * IMP_BETAS = gsl_matrix_calloc(p,1);
	for(i=0; i<p; i++) {gsl_matrix_set(IMP_BETAS, i, 0, imp_betas[i]);}//populate with initial values
	gsl_matrix * SCALED_IMP_BETAS = gsl_matrix_calloc(p,1);//X_SCALE_FACTOR%*%IMP_BETAS
	gsl_matrix * MEAN_SCALED_IMP_BETAS = gsl_matrix_calloc(p,1);//mean matrix (vector) of betas
	gsl_matrix * CHOL_PRECISION_SCALED_IMP_BETAS = gsl_matrix_calloc(p,p);//Lower triangular L such that Variance(imp_betas) = (L^-1)^T %*% (L^-1)
	gsl_matrix * INV_CHOL_PRECISION_SCALED_IMP_BETAS = gsl_matrix_calloc(p,p);//L^(-1)
	gsl_matrix * RANDOM_VALS_SCALED_IMP_BETAS = gsl_matrix_calloc(p,1);//stochastic component
	gsl_matrix * CROSSPROD_S_XA_YA = gsl_matrix_calloc(p,1);//t(s_xa)%*%ya
	gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,S_XA,YA,0.,CROSSPROD_S_XA_YA);
	gsl_matrix * CROSSPROD_S_XA_S_XA = gsl_matrix_calloc(p,p);//t(s_xa)%*%s_xa
	gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,S_XA,S_XA,0.,CROSSPROD_S_XA_S_XA);
	gsl_matrix * CROSSPROD_IMP_S_XB_YB = gsl_matrix_calloc(p,1);//t(s_xb)%*%yb
	gsl_matrix * CROSSPROD_IMP_S_XB_IMP_S_XB = gsl_matrix_calloc(p,p);//t(s_xb)%*%s_xb
	gsl_matrix * COLSUMS_S_XA = gsl_matrix_calloc(p,1);//column vector of column sums of s_xa
	column_sums(na,p,S_XA,COLSUMS_S_XA);
	gsl_matrix * COLSUMS_IMP_S_XB = gsl_matrix_calloc(p,1);//column vector of column sums of imp_s_xb

	//beta0
	gsl_matrix * REG_RESID_A = gsl_matrix_calloc(na,1);//regression residuals: ya - xa%*%betas	
	gsl_matrix * REG_RESID_B = gsl_matrix_calloc(nb,1);//regression residuals: yb - imp_xb%*%betas	
	gsl_matrix * SCALED_OFFSET = gsl_matrix_calloc(p,1);//location_factor^T%*%imp_betas
	gsl_matrix_memcpy(SCALED_OFFSET,IMP_BETAS);
	gsl_matrix_mul_elements(SCALED_OFFSET,X_LOCATION_FACTOR);
	double scaled_imp_beta0 = *imp_beta0 + element_wise_sum(p,1,SCALED_OFFSET,1.);
	double sum_REG_RESID_A = 0.;
	double sum_REG_RESID_B = 0.;

	//sigmasq
	double dotprod_SCALED_IMP_BETAS = 0.;//(SCALED_IMP_BETAS)^T%*%SCALED_IMP_BETAS
	double dotprod_REG_RESID_A = 0.;//(ya-xa%*%imp_betas)^T%*%(ya-xa%*%imp_betas)
	double dotprod_REG_RESID_B= 0.;//(yb-xb%*%imp_betas)^T%*%(yb-xb%*%imp_betas)
	double as_double_n = (double)(na+nb);
	double shape_imp_sigmasq = as_double_n / 2.;//shape parameter for posterior distribution
	if(*imp_lambda_betas > 0.) {shape_imp_sigmasq += (double)p / 2.;}
	double scale_imp_sigmasq;

	//psi
	gsl_matrix * IMP_PSI = gsl_matrix_calloc(p,1);
	for(i=0; i<p; i++) {gsl_matrix_set(IMP_PSI, i, 0, imp_psi[i]);}//populate with initial values
	gsl_matrix * MEAN_IMP_PSI = gsl_matrix_calloc(p,1);
	gsl_matrix * COLSUMS_XA = gsl_matrix_calloc(p,1);//column vector of column sums of xa
	column_sums(na,p,XA,COLSUMS_XA);
	gsl_matrix * COLSUMS_WA = gsl_matrix_calloc(p,1);//column vector of column sums of wa
	column_sums(na,p,WA,COLSUMS_WA);
	gsl_matrix * COLSUMS_IMP_XB = gsl_matrix_calloc(p,1);//column vector of column sums of imp_xb
	gsl_matrix * COLSUMS_WB = gsl_matrix_calloc(p,1);//column vector of column sums of wb
	column_sums(nb,p,WB,COLSUMS_WB);
	double as_double_n_times_p = (double)((na+nb)*p);

	//nu
	gsl_matrix * DIAG_IMP_NU = gsl_matrix_calloc(p,p);
	for(i=0; i<p; i++) {gsl_matrix_set(DIAG_IMP_NU, i, i, imp_nu[i]);}//populate with initial values
	gsl_matrix * NUMERATOR_MEAN_IMP_NU = gsl_matrix_calloc(p,1);
	gsl_matrix * DENOMINATOR_MEAN_IMP_NU = gsl_matrix_calloc(p,1);	
	gsl_matrix * CROSSPROD_XA_XA = gsl_matrix_calloc(p,p);//(xa)^T%*%xa
	gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,XA,XA,0.,CROSSPROD_XA_XA);
	gsl_matrix * DIAG_CROSSPROD_XA_XA = gsl_matrix_calloc(p,1);
	for(i=0; i<p; i++) {gsl_matrix_set(DIAG_CROSSPROD_XA_XA,i,0,gsl_matrix_get(CROSSPROD_XA_XA,i,i));}
	gsl_matrix * CROSSPROD_XA_WA = gsl_matrix_calloc(p,p);//(xa)^T%*%wa
	gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,XA,WA,0.,CROSSPROD_XA_WA);
	gsl_matrix * DIAG_CROSSPROD_XA_WA = gsl_matrix_calloc(p,1);
	for(i=0; i<p; i++) {gsl_matrix_set(DIAG_CROSSPROD_XA_WA,i,0,gsl_matrix_get(CROSSPROD_XA_WA,i,i));}
	gsl_matrix * CROSSPROD_IMP_XB_IMP_XB = gsl_matrix_calloc(p,p);//(xb)^T%*%xb
	gsl_matrix * DIAG_CROSSPROD_IMP_XB_IMP_XB = gsl_matrix_calloc(p,1);
	gsl_matrix * CROSSPROD_IMP_XB_WB = gsl_matrix_calloc(p,p);//(xb)^T%*%wb
	gsl_matrix * DIAG_CROSSPROD_IMP_XB_WB = gsl_matrix_calloc(p,1);


	//tausq
	gsl_matrix * ME_RESID_A = gsl_matrix_calloc(na,p);//measurement error residuals: wa-imp_nu*xa
	gsl_matrix * ME_RESID_B = gsl_matrix_calloc(nb,p);//measurement error residuals: wb-imp_nu*imp_xb
	double shape_imp_tausq = as_double_n_times_p / 2.;
	double scale_imp_tausq;

	//mean_x
	gsl_matrix * IMP_MEAN_X = gsl_matrix_calloc(p,1);
	for(i=0; i<p; i++) {gsl_matrix_set(IMP_MEAN_X, i, 0, imp_mean_x[i]);}//populate with initial values
	gsl_matrix * RANDOM_VALS_IMP_MEAN_X = gsl_matrix_calloc(p,1);//stochastic component
	gsl_matrix * CHOL_IMP_PRECISION_X = gsl_matrix_calloc(p,p);
	gsl_matrix * INV_CHOL_IMP_PRECISION_X = gsl_matrix_calloc(p,p);

	//precision_x
	gsl_matrix * IMP_PRECISION_X = gsl_matrix_calloc(p,p);
	for(i=0; i<p; i++) for(j=0; j<p; j++) {gsl_matrix_set(IMP_PRECISION_X, i, j, imp_precision_x[i*p+j]);}//populate with initial values
	int df_IMP_PRECISION_X;
	gsl_matrix * CHOL_INVSCALE_IMP_PRECISION_X = gsl_matrix_calloc(p,p);
	gsl_matrix * INV_CHOL_INVSCALE_IMP_PRECISION_X = gsl_matrix_calloc(p,p);
	double lndet_IMP_PRECISION_X;

	//lambda_betas
	int n_ebiter_betas = 10;
	int m1 = 0;
	double store_dotprod_SCALED_IMP_BETAS_over_imp_sigmasq = 0.;

	//lambda_precision_x
	gsl_matrix * IMP_LAMBDA_PRECISION_X = gsl_matrix_calloc(p,p);
	for(i=0; i<p; i++) for(j=0; j<p; j++) {gsl_matrix_set(IMP_LAMBDA_PRECISION_X, i, j, imp_lambda_precision_x[i*p+j]);}//populate with initial values
	int n_ebiter_precision_X = (int)(ceil(p/4));
	int m2 = 0;
	gsl_matrix * STORE_PRECISION_X = gsl_matrix_calloc(p,p);
	double store_log_det_PRECISION_X = 0.;	
	double trace_IMP_LAMBDA_PRECISION_X = matrix_trace(p, IMP_LAMBDA_PRECISION_X);
	int prior_df_grid[10*p];
	for(i=0;i<10*p;i++) {
		prior_df_grid[i] = *precision_x_prior_df-(5*p)+(i+1);	
	}

	//Return values
	*return_beta0 = 0.;
	gsl_matrix * QUADLOSS_RETURN_BETAS = gsl_matrix_calloc(p,1);
	gsl_matrix * PREDLOSS_RETURN_BETAS = gsl_matrix_calloc(p,1);
	gsl_matrix * CURRENT_EXP_XXTRANS = gsl_matrix_calloc(p,p);
	gsl_matrix * RETURN_EXP_XXTRANS = gsl_matrix_calloc(p,p);
	gsl_matrix * RETURN_EXP_XXTRANS_BETAS = gsl_matrix_calloc(p,1);
	gsl_matrix * RETURN_ALL_BETAS = gsl_matrix_calloc(mcmc_niter,p);
	gsl_matrix * RETURN_ALL_LAMBDA_PRECISION_X = gsl_matrix_calloc(mcmc_niter,p);

	//
	
	for(k = 0; k<(mcmc_burnin+mcmc_niter);k++) {

		//////////////////////
		//Imputation of xb//
		/////////////////////

		//First draw stochastic component
		for (i=0; i<nb; i++) for(j=0; j<p; j++) {gsl_matrix_set (RANDOM_VALS_IMP_XB, i, j, gsl_ran_ugaussian(ran_num_gen));}
			
		//Now determine the covariance matrix (via the inverse of the cholesky decomposition of the precision)
		// precision <- nu^2
		gsl_matrix_memcpy(CHOL_PRECISION_IMP_XB,DIAG_IMP_NU);
		gsl_matrix_mul_elements(CHOL_PRECISION_IMP_XB,DIAG_IMP_NU);
		// precision <- beta*beta^T/sigma2 + precision/tausq
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,(1./ *imp_sigmasq),IMP_BETAS,IMP_BETAS,(1. / *imp_tausq),CHOL_PRECISION_IMP_XB);
		// precision <- precision + precision_x
		gsl_matrix_add(CHOL_PRECISION_IMP_XB,IMP_PRECISION_X);
		//take sqrt, will use only lower triangular values (L)
		*error_code = gsl_linalg_cholesky_decomp(CHOL_PRECISION_IMP_XB);
		if(*error_code == GSL_EDOM) {*error_code=1; break;}
		//now invert (L^(-1)
		gsl_matrix_set_identity(INV_CHOL_PRECISION_IMP_XB);
		gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1., CHOL_PRECISION_IMP_XB, INV_CHOL_PRECISION_IMP_XB);


		//Now determine the mean of xb
		//mean <- (wb-1%*%t(IMP_PSI))%*%DIAG_IMP_NU/imp_tausq
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,ONES_LENGTH_NB,IMP_PSI,0,MEAN_IMP_XB);
		gsl_matrix_add(MEAN_IMP_XB,WB);
		gsl_blas_dtrmm(CblasRight, CblasLower,CblasNoTrans,CblasNonUnit,(1. / *imp_tausq),DIAG_IMP_NU,MEAN_IMP_XB);
		//mean <- yb*beta^T/sigma2 + mean
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,(1./ *imp_sigmasq),YB,IMP_BETAS,1.,MEAN_IMP_XB);
		//mean <- mean - beta0/sigma2*beta
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-(*imp_beta0 / *imp_sigmasq),ONES_LENGTH_NB,IMP_BETAS,1.,MEAN_IMP_XB);
		//mean <- mean + 1*mu_x^T*precision_x
		gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,IMP_MEAN_X,IMP_PRECISION_X,0.,CROSSPROD_IMP_MEAN_X_IMP_PRECISION_X);
		gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,1.,ONES_LENGTH_NB,CROSSPROD_IMP_MEAN_X_IMP_PRECISION_X,1.,MEAN_IMP_XB);
		//mean <- mean*(L^-1)^T
		gsl_blas_dtrmm (CblasRight, CblasLower,CblasTrans,CblasNonUnit,1.,INV_CHOL_PRECISION_IMP_XB,MEAN_IMP_XB);
		//mean <- mean*(L^-1)
		gsl_blas_dtrmm (CblasRight, CblasLower,CblasNoTrans,CblasNonUnit,1.,INV_CHOL_PRECISION_IMP_XB,MEAN_IMP_XB);

		// put xb <- mean + random_vals * sqrt(variance))
		gsl_matrix_memcpy(IMP_XB,MEAN_IMP_XB);
		gsl_blas_dtrmm(CblasRight,CblasLower,CblasNoTrans,CblasNonUnit,1.,INV_CHOL_PRECISION_IMP_XB,RANDOM_VALS_IMP_XB);
		gsl_matrix_add(IMP_XB,RANDOM_VALS_IMP_XB);


		//Store cross products
		gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,IMP_XB,IMP_XB,0.,CROSSPROD_IMP_XB_IMP_XB);

		//Store colsums
		column_sums(nb,p,IMP_XB,COLSUMS_IMP_XB);

		//calculate IMP_S_XB
		gsl_matrix_memcpy(IMP_S_XB,IMP_XB);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,ONES_LENGTH_NB,X_LOCATION_FACTOR,1.,IMP_S_XB);
		gsl_blas_dtrmm (CblasRight, CblasLower,CblasNoTrans,CblasNonUnit,1.,X_INV_SCALE_FACTOR,IMP_S_XB);

		//////////////////////
		//Imputation of scaled_betas//
		/////////////////////

		column_sums(nb,p,IMP_S_XB,COLSUMS_IMP_S_XB);
		gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,IMP_S_XB,IMP_S_XB,0.,CROSSPROD_IMP_S_XB_IMP_S_XB);
		gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,IMP_S_XB,YB,0.,CROSSPROD_IMP_S_XB_YB);

		//First draw stochastic component
		for(i=0; i<p; i++) {gsl_matrix_set (RANDOM_VALS_SCALED_IMP_BETAS, i, 0, gsl_ran_ugaussian(ran_num_gen));}
		
		//Now the covariance matrix (via the precision)
		// put precision <- imp_lambda_betas*Ip+(s_xa)^T%*%(s_xa)+(imp_s_xb)^T%*%imp_s_xb
		gsl_matrix_set_identity(CHOL_PRECISION_SCALED_IMP_BETAS);
		gsl_matrix_scale(CHOL_PRECISION_SCALED_IMP_BETAS,*imp_lambda_betas);
		gsl_matrix_add(CHOL_PRECISION_SCALED_IMP_BETAS,CROSSPROD_S_XA_S_XA);	
		gsl_matrix_add(CHOL_PRECISION_SCALED_IMP_BETAS,CROSSPROD_IMP_S_XB_IMP_S_XB);


		//take sqrt, will use only lower triangular values (L)
		*error_code = gsl_linalg_cholesky_decomp(CHOL_PRECISION_SCALED_IMP_BETAS);
		if(*error_code == GSL_EDOM) {*error_code=2; break;}
		//now invert (L^(-1)
		gsl_matrix_set_identity(INV_CHOL_PRECISION_SCALED_IMP_BETAS);
		gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1., CHOL_PRECISION_SCALED_IMP_BETAS, INV_CHOL_PRECISION_SCALED_IMP_BETAS);

		//Now the mean
		//mean <- -beta0*(colsums(s_xa)+colsums(imp_s_xb)) + (s_xa)^T%*%ya + (imp_s_xb)^T%*%yb 
		gsl_matrix_memcpy(MEAN_SCALED_IMP_BETAS,COLSUMS_S_XA);
		gsl_matrix_add(MEAN_SCALED_IMP_BETAS,COLSUMS_IMP_S_XB);
		gsl_matrix_scale(MEAN_SCALED_IMP_BETAS,-(scaled_imp_beta0));
		gsl_matrix_add(MEAN_SCALED_IMP_BETAS,CROSSPROD_S_XA_YA);
		gsl_matrix_add(MEAN_SCALED_IMP_BETAS,CROSSPROD_IMP_S_XB_YB);
		//mean <- (L^-1)*mean
		gsl_blas_dtrmm (CblasLeft, CblasLower,CblasNoTrans,CblasNonUnit,1.,INV_CHOL_PRECISION_SCALED_IMP_BETAS,MEAN_SCALED_IMP_BETAS);
		//mean <- (L^-1)^T*mean
		gsl_blas_dtrmm (CblasLeft, CblasLower,CblasTrans,CblasNonUnit,1.,INV_CHOL_PRECISION_SCALED_IMP_BETAS,MEAN_SCALED_IMP_BETAS);

		// put imp_betas <- mean + sqrt(sigmasq)*sqrt(variance)*random_vals )
		gsl_matrix_memcpy(SCALED_IMP_BETAS,MEAN_SCALED_IMP_BETAS);
		gsl_blas_dtrmm(CblasLeft,CblasLower,CblasTrans,CblasNonUnit,sqrt(*imp_sigmasq),INV_CHOL_PRECISION_SCALED_IMP_BETAS,RANDOM_VALS_SCALED_IMP_BETAS);
		gsl_matrix_add(SCALED_IMP_BETAS,RANDOM_VALS_SCALED_IMP_BETAS);

		//imp_betas (original scale)
		gsl_matrix_memcpy(IMP_BETAS,SCALED_IMP_BETAS);
		gsl_blas_dtrmm (CblasLeft, CblasLower,CblasNoTrans,CblasNonUnit,1.,X_INV_SCALE_FACTOR,IMP_BETAS);
		
		//////////////////////
		//Imputation of beta0
		/////////////////////

		gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,-1.,XA,IMP_BETAS,0.,REG_RESID_A);
		gsl_matrix_add (REG_RESID_A, YA);
		sum_REG_RESID_A = element_wise_sum(na,1,REG_RESID_A,1.);

		gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,-1.,IMP_XB,IMP_BETAS,0.,REG_RESID_B);
		gsl_matrix_add (REG_RESID_B, YB);
		sum_REG_RESID_B = element_wise_sum(nb,1,REG_RESID_B,1.);

		*imp_beta0 = (sum_REG_RESID_A+sum_REG_RESID_B)/as_double_n + sqrt(*imp_sigmasq/as_double_n)*gsl_ran_ugaussian(ran_num_gen);
		gsl_matrix_memcpy(SCALED_OFFSET,IMP_BETAS);
		gsl_matrix_mul_elements(SCALED_OFFSET,X_LOCATION_FACTOR);
		scaled_imp_beta0 = *imp_beta0 + element_wise_sum(p,1,SCALED_OFFSET,1.);

		//////////////////////
		//Imputation of sigmasq//
		/////////////////////

		dotprod_SCALED_IMP_BETAS = element_wise_sum(p,1,SCALED_IMP_BETAS,2.);

		gsl_matrix_add_constant (REG_RESID_A, -(*imp_beta0));
		dotprod_REG_RESID_A = element_wise_sum(na,1,REG_RESID_A,2.);

		gsl_matrix_add_constant (REG_RESID_B, -(*imp_beta0));
		dotprod_REG_RESID_B= element_wise_sum(nb,1,REG_RESID_B,2.);

		scale_imp_sigmasq = 2 / ((*imp_lambda_betas) * (dotprod_SCALED_IMP_BETAS) + dotprod_REG_RESID_A + dotprod_REG_RESID_B);
		*imp_sigmasq = 1./gsl_ran_gamma(ran_num_gen,shape_imp_sigmasq,scale_imp_sigmasq);
	
		//////////////////////
		//Imputation of psi//
		/////////////////////
		
		column_sums(nb,p,IMP_XB,COLSUMS_IMP_XB);
		gsl_matrix_memcpy(MEAN_IMP_PSI,COLSUMS_XA);
		gsl_matrix_add(MEAN_IMP_PSI,COLSUMS_IMP_XB);
		gsl_blas_dtrmm(CblasLeft,CblasUpper,CblasNoTrans,CblasNonUnit,-1.,DIAG_IMP_NU,MEAN_IMP_PSI);
		gsl_matrix_add(MEAN_IMP_PSI,COLSUMS_WA);
		gsl_matrix_add(MEAN_IMP_PSI,COLSUMS_WB);
		gsl_matrix_scale(MEAN_IMP_PSI,(1./(double)(na+nb)));
		for(i=0; i<p; i++) {
			gsl_matrix_set(IMP_PSI,i,0, gsl_matrix_get(MEAN_IMP_PSI,i,0) + sqrt((*imp_tausq))/sqrt((double)(na+nb)) * gsl_ran_ugaussian(ran_num_gen));
		}


		//////////////////////
		//Imputation of nu//
		/////////////////////

		gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,IMP_XB,IMP_XB,0.,CROSSPROD_IMP_XB_IMP_XB);
		for(i=0; i<p; i++) {gsl_matrix_set(DIAG_CROSSPROD_IMP_XB_IMP_XB,i,0,gsl_matrix_get(CROSSPROD_IMP_XB_IMP_XB,i,i));}
		gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,IMP_XB,WB,0.,CROSSPROD_IMP_XB_WB);			
		for(i=0; i<p; i++) {gsl_matrix_set(DIAG_CROSSPROD_IMP_XB_WB,i,0,gsl_matrix_get(CROSSPROD_IMP_XB_WB,i,i));}
				
		gsl_matrix_memcpy(NUMERATOR_MEAN_IMP_NU,COLSUMS_XA);
		gsl_matrix_add(NUMERATOR_MEAN_IMP_NU,COLSUMS_IMP_XB);
		gsl_matrix_scale(NUMERATOR_MEAN_IMP_NU,-1.);
		gsl_matrix_mul_elements(NUMERATOR_MEAN_IMP_NU,IMP_PSI);
		gsl_matrix_add(NUMERATOR_MEAN_IMP_NU,DIAG_CROSSPROD_XA_WA);
		gsl_matrix_add(NUMERATOR_MEAN_IMP_NU,DIAG_CROSSPROD_IMP_XB_WB);

		gsl_matrix_memcpy(DENOMINATOR_MEAN_IMP_NU,DIAG_CROSSPROD_XA_XA);
		gsl_matrix_add(DENOMINATOR_MEAN_IMP_NU,DIAG_CROSSPROD_IMP_XB_IMP_XB);

		gsl_matrix_div_elements(NUMERATOR_MEAN_IMP_NU,DENOMINATOR_MEAN_IMP_NU);
		for(i=0; i<p; i++) {
			gsl_matrix_set(DIAG_IMP_NU,i,i,gsl_matrix_get(NUMERATOR_MEAN_IMP_NU,i,0) +sqrt((*imp_tausq)/gsl_matrix_get(DENOMINATOR_MEAN_IMP_NU,i,0))* gsl_ran_ugaussian(ran_num_gen));
		}			

		//////////////////////
		//Imputation of tausq//
		/////////////////////

		gsl_matrix_memcpy(ME_RESID_A,XA);
		gsl_blas_dtrmm(CblasRight,CblasUpper,CblasNoTrans,CblasNonUnit,-1.,DIAG_IMP_NU,ME_RESID_A);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,ONES_LENGTH_NA,IMP_PSI,1.,ME_RESID_A);
		gsl_matrix_add(ME_RESID_A,WA);

		gsl_matrix_memcpy(ME_RESID_B,IMP_XB);
		gsl_blas_dtrmm(CblasRight,CblasUpper,CblasNoTrans,CblasNonUnit,-1.,DIAG_IMP_NU,ME_RESID_B);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,ONES_LENGTH_NB,IMP_PSI,1.,ME_RESID_B);
		gsl_matrix_add(ME_RESID_B,WB);
		scale_imp_tausq = 2. / (element_wise_sum(na,p,ME_RESID_A,2.)+element_wise_sum(nb,p,ME_RESID_B,2.));
		*imp_tausq = 1./gsl_ran_gamma(ran_num_gen,shape_imp_tausq,scale_imp_tausq);
	
		//////////////////////
		//Imputation of precision_x//
		/////////////////////
		
		df_IMP_PRECISION_X = (int)(na + nb);
		if(trace_IMP_LAMBDA_PRECISION_X>0.) {df_IMP_PRECISION_X += *precision_x_prior_df;}

		// put CHOL_INVSCALE_IMP_PRECISION_X <- (xa - 1_na %*% mean_x^T)^T %*% (xa - 1_na %*% mean_x^T)
		// by adding each term in the cross product			
		gsl_matrix_memcpy(CHOL_INVSCALE_IMP_PRECISION_X,CROSSPROD_XA_XA);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,IMP_MEAN_X,COLSUMS_XA,1.,CHOL_INVSCALE_IMP_PRECISION_X);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,COLSUMS_XA,IMP_MEAN_X,1.,CHOL_INVSCALE_IMP_PRECISION_X);

		// put CHOL_INVSCALE_IMP_PRECISION_X <- CHOL_INVSCALE_IMP_PRECISION_X + (imp_xb - 1_nb %*% mean_x^T)^T %*% (imp_xb - 1_nb %*% mean_x^T)
		//by adding each term in the cross product
		gsl_matrix_add(CHOL_INVSCALE_IMP_PRECISION_X,CROSSPROD_IMP_XB_IMP_XB);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,IMP_MEAN_X,COLSUMS_IMP_XB,1.,CHOL_INVSCALE_IMP_PRECISION_X);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,COLSUMS_IMP_XB,IMP_MEAN_X,1.,CHOL_INVSCALE_IMP_PRECISION_X);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,as_double_n,IMP_MEAN_X,IMP_MEAN_X,1.,CHOL_INVSCALE_IMP_PRECISION_X);

		// put CHOL_INVSCALE_IMP_PRECISION_X <- CHOL_INVSCALE_IMP_PRECISION_X + IMP_LAMBDA_PRECISION_X
		gsl_matrix_add(CHOL_INVSCALE_IMP_PRECISION_X,IMP_LAMBDA_PRECISION_X);

		//take sqrt, will use only lower triangular values (L)
		*error_code = gsl_linalg_cholesky_decomp(CHOL_INVSCALE_IMP_PRECISION_X);
		if(*error_code == GSL_EDOM) {*error_code=3; break;}
		//now invert (L^(-1)
		gsl_matrix_set_identity(INV_CHOL_INVSCALE_IMP_PRECISION_X);
		gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1., CHOL_INVSCALE_IMP_PRECISION_X, INV_CHOL_INVSCALE_IMP_PRECISION_X);
		
		//Pass this to wishart to draw a random wishart
		lndet_IMP_PRECISION_X = wishart(df_IMP_PRECISION_X, INV_CHOL_INVSCALE_IMP_PRECISION_X, p, ran_num_gen, IMP_PRECISION_X);

		//////////////////////
		//Imputation of mean_x//
		/////////////////////

		//First draw stochastic component
		for(i=0; i<p; i++) {gsl_matrix_set (RANDOM_VALS_IMP_MEAN_X, i, 0, gsl_ran_ugaussian(ran_num_gen));}
		
		//take sqrt of precision_x, will use only lower triangular values (L)
		gsl_matrix_memcpy(CHOL_IMP_PRECISION_X,IMP_PRECISION_X);
		*error_code = gsl_linalg_cholesky_decomp(CHOL_IMP_PRECISION_X);
		if(*error_code == GSL_EDOM) {*error_code=4; break;}		
		//now invert (L^-1)
		gsl_matrix_set_identity(INV_CHOL_IMP_PRECISION_X);
		gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1., CHOL_IMP_PRECISION_X, INV_CHOL_IMP_PRECISION_X);	
	
		//put imp_mean_x <- (colsums(xa)+colsums(imp_xb))/(na+nb)
		gsl_matrix_memcpy(IMP_MEAN_X,COLSUMS_XA);
		gsl_matrix_add(IMP_MEAN_X,COLSUMS_IMP_XB);
		gsl_matrix_scale(IMP_MEAN_X,1./as_double_n);

		//Now add the stochastic component (ie put imp_mean_x <- imp_mean_x + sqrt(variance)*random_seed )
		gsl_blas_dtrmm(CblasLeft,CblasLower,CblasTrans,CblasNonUnit,1./sqrt(as_double_n),INV_CHOL_IMP_PRECISION_X,RANDOM_VALS_IMP_MEAN_X);
		gsl_matrix_add(IMP_MEAN_X,RANDOM_VALS_IMP_MEAN_X);


		//////////////////////
		//Imputation of lambda_betas//
		/////////////////////
 		if(lambda_betas_update==1) {//empirical bayes update
			m1 += 1;
			store_dotprod_SCALED_IMP_BETAS_over_imp_sigmasq += dotprod_SCALED_IMP_BETAS/(*imp_sigmasq);
			if(m1==n_ebiter_betas) {
				if(store_dotprod_SCALED_IMP_BETAS_over_imp_sigmasq==0.) {
					*error_code=5; 
					break;
				} else {
					*imp_lambda_betas = (double)((n_ebiter_betas * p)/store_dotprod_SCALED_IMP_BETAS_over_imp_sigmasq);
				}
				store_dotprod_SCALED_IMP_BETAS_over_imp_sigmasq = 0.;
				m1 = 0;
				if(k>=mcmc_burnin) {n_ebiter_betas=50;}
			}
		} else if(lambda_betas_update==2) {//hierarchical update
			*imp_lambda_betas = gsl_ran_gamma(ran_num_gen,(double)(p/2)+lambda_betas_prior_shape,1./((dotprod_SCALED_IMP_BETAS/(*imp_sigmasq))/2.+lambda_betas_prior_rate));
		}

		//////////////////////
		//Imputation of lambda_precision_x//
		/////////////////////
		if(lambda_precision_x_update==1) {
			m2 += 1;
			if(constrain_lambda_precision_x==1) {
				gsl_matrix_add_diagonals(p,STORE_PRECISION_X,IMP_PRECISION_X);
			} else if(constrain_lambda_precision_x==2) {
				gsl_matrix_add_constant_diagonal_only(p,STORE_PRECISION_X,matrix_trace(p, IMP_PRECISION_X)/(double)(p));
			}
			store_log_det_PRECISION_X += lndet_IMP_PRECISION_X;
			if(m2==n_ebiter_precision_X) {
				gsl_matrix_set_identity(IMP_LAMBDA_PRECISION_X);
				*error_code = gsl_linalg_cholesky_decomp(STORE_PRECISION_X);
				if(*error_code == GSL_EDOM) {*error_code=6; break;}
				gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1., STORE_PRECISION_X, IMP_LAMBDA_PRECISION_X);	
				gsl_blas_dtrsm(CblasLeft, CblasLower, CblasTrans, CblasNonUnit, (double)(*precision_x_prior_df*n_ebiter_precision_X), STORE_PRECISION_X, IMP_LAMBDA_PRECISION_X);
				gsl_matrix_set_zero(STORE_PRECISION_X);
				trace_IMP_LAMBDA_PRECISION_X = matrix_trace(p, IMP_LAMBDA_PRECISION_X);
				m2 = 0;
				if(k>=1000) {n_ebiter_precision_X=3*p;}
			}
		} else if(lambda_precision_x_update==2) {
			m2 += 1;
			store_log_det_PRECISION_X += lndet_IMP_PRECISION_X;
			if(m2==n_ebiter_precision_X) {
				*precision_x_prior_df = best_df(prior_df_grid,10*p,p*log(2)+gsl_linalg_LU_lndet(IMP_LAMBDA_PRECISION_X)+(store_log_det_PRECISION_X/n_ebiter_precision_X),p);
				for(i=0;i<10*p;i++) {
					prior_df_grid[i] = *precision_x_prior_df-(5*p)+(i+1);
				}
				store_log_det_PRECISION_X = 0.;
				m2 = 0;
				if(k>=1000) {n_ebiter_precision_X=3*p;}
			}
		}

		if(k>=mcmc_burnin) {
			gsl_matrix_add(QUADLOSS_RETURN_BETAS,IMP_BETAS);
			gsl_matrix_set_identity(CURRENT_EXP_XXTRANS);
			gsl_blas_dtrmm(CblasRight, CblasLower, CblasTrans, CblasNonUnit, 1., INV_CHOL_IMP_PRECISION_X, CURRENT_EXP_XXTRANS);		
			gsl_blas_dtrmm(CblasRight, CblasLower, CblasNoTrans, CblasNonUnit, 1., INV_CHOL_IMP_PRECISION_X, CURRENT_EXP_XXTRANS);		
			gsl_blas_dgemm(CblasNoTrans,CblasTrans,1.,IMP_MEAN_X,IMP_MEAN_X,1.,CURRENT_EXP_XXTRANS);
			gsl_matrix_add(RETURN_EXP_XXTRANS,CURRENT_EXP_XXTRANS);
			gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,1.,CURRENT_EXP_XXTRANS,IMP_BETAS,1.,RETURN_EXP_XXTRANS_BETAS);	
			
			*return_beta0 += (*imp_beta0) / (double)(mcmc_niter);

			gsl_vector_view VIEW_AS_VECTOR_IMP_BETAS = gsl_matrix_column (IMP_BETAS, 0);
			gsl_matrix_set_row(RETURN_ALL_BETAS,k-mcmc_burnin, &VIEW_AS_VECTOR_IMP_BETAS.vector);
			store_beta0[k-mcmc_burnin] = (*imp_beta0);
			store_sigmasq[k-mcmc_burnin] = (*imp_sigmasq);
			store_tausq[k-mcmc_burnin] = (*imp_tausq);
			store_lambda_betas[k-mcmc_burnin] = (*imp_lambda_betas);
			store_trace_lambda_precision_x[k-mcmc_burnin] = trace_IMP_LAMBDA_PRECISION_X;
			store_precision_x_prior_df[k-mcmc_burnin] = *precision_x_prior_df;
			gsl_vector_view VIEW_AS_VECTOR_DIAGONAL_IMP_LAMBDA_PRECISION_X = gsl_matrix_diagonal (IMP_LAMBDA_PRECISION_X);
			gsl_matrix_set_row(RETURN_ALL_LAMBDA_PRECISION_X,k-mcmc_burnin, &VIEW_AS_VECTOR_DIAGONAL_IMP_LAMBDA_PRECISION_X.vector);
		}

	}

	////////////////
	//Clean up
	///////////////
	

	gsl_rng_free (ran_num_gen);

	//data
	gsl_matrix_free (YA );
	gsl_matrix_free (YB );
	gsl_matrix_free (XA );		
	gsl_matrix_free (WA );
	gsl_matrix_free (WB );
	gsl_matrix_free (IMP_XB );

	//scaled versions
	gsl_matrix_free (X_INV_SCALE_FACTOR);
	gsl_matrix_free (X_LOCATION_FACTOR);
	gsl_matrix_free (ONES_LENGTH_NA);
	gsl_matrix_free (ONES_LENGTH_NB);
	gsl_matrix_free (S_XA );
	gsl_matrix_free (IMP_S_XB );

	//xb
	gsl_matrix_free (MEAN_IMP_XB );
	gsl_matrix_free (CHOL_PRECISION_IMP_XB );
	gsl_matrix_free (INV_CHOL_PRECISION_IMP_XB );
	gsl_matrix_free (RANDOM_VALS_IMP_XB );
	gsl_matrix_free (CROSSPROD_IMP_MEAN_X_IMP_PRECISION_X );

	//betas
	gsl_matrix_free (IMP_BETAS );
	gsl_matrix_free (SCALED_IMP_BETAS );
	gsl_matrix_free (MEAN_SCALED_IMP_BETAS );
	gsl_matrix_free (CHOL_PRECISION_SCALED_IMP_BETAS );
	gsl_matrix_free (INV_CHOL_PRECISION_SCALED_IMP_BETAS );
	gsl_matrix_free (RANDOM_VALS_SCALED_IMP_BETAS );
	gsl_matrix_free (CROSSPROD_S_XA_YA);
	gsl_matrix_free (CROSSPROD_S_XA_S_XA);
	gsl_matrix_free (CROSSPROD_IMP_S_XB_YB );
	gsl_matrix_free (CROSSPROD_IMP_S_XB_IMP_S_XB );
	gsl_matrix_free (COLSUMS_IMP_S_XB );
	gsl_matrix_free (COLSUMS_S_XA);

	//beta0
	gsl_matrix_free (REG_RESID_A );
	gsl_matrix_free (REG_RESID_B );
	gsl_matrix_free (SCALED_OFFSET );

	//psi
	gsl_matrix_free (IMP_PSI);
	gsl_matrix_free (MEAN_IMP_PSI);	
	gsl_matrix_free (COLSUMS_XA);
	gsl_matrix_free (COLSUMS_WA);	
	gsl_matrix_free (COLSUMS_IMP_XB);
	gsl_matrix_free (COLSUMS_WB);

	//nu
	gsl_matrix_free (DIAG_IMP_NU);
	gsl_matrix_free (NUMERATOR_MEAN_IMP_NU);
	gsl_matrix_free (DENOMINATOR_MEAN_IMP_NU);
	gsl_matrix_free (CROSSPROD_XA_XA);
	gsl_matrix_free (DIAG_CROSSPROD_XA_XA);
	gsl_matrix_free (CROSSPROD_XA_WA);
	gsl_matrix_free (DIAG_CROSSPROD_XA_WA);
	gsl_matrix_free (CROSSPROD_IMP_XB_IMP_XB);
	gsl_matrix_free (DIAG_CROSSPROD_IMP_XB_IMP_XB);
	gsl_matrix_free (CROSSPROD_IMP_XB_WB);
	gsl_matrix_free (DIAG_CROSSPROD_IMP_XB_WB);

	//tau2
	gsl_matrix_free (ME_RESID_A );
	gsl_matrix_free (ME_RESID_B );

	//mean_x
	gsl_matrix_free (IMP_MEAN_X);
	gsl_matrix_free (RANDOM_VALS_IMP_MEAN_X );
	gsl_matrix_free (CHOL_IMP_PRECISION_X );
	gsl_matrix_free (INV_CHOL_IMP_PRECISION_X );

	//precision_x
	gsl_matrix_free (IMP_PRECISION_X);
	gsl_matrix_free (CHOL_INVSCALE_IMP_PRECISION_X  );
	gsl_matrix_free (INV_CHOL_INVSCALE_IMP_PRECISION_X );

	//lambda_precision_x
	gsl_matrix_free (IMP_LAMBDA_PRECISION_X);
	gsl_matrix_free (STORE_PRECISION_X);

	if(*error_code==0) {

		gsl_matrix_scale(QUADLOSS_RETURN_BETAS,1./(double)(mcmc_niter));

		gsl_matrix_memcpy(PREDLOSS_RETURN_BETAS,RETURN_EXP_XXTRANS_BETAS);
		gsl_linalg_cholesky_decomp(RETURN_EXP_XXTRANS);
		gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1., RETURN_EXP_XXTRANS, PREDLOSS_RETURN_BETAS);	
		gsl_blas_dtrsm(CblasLeft, CblasLower, CblasTrans, CblasNonUnit, 1., RETURN_EXP_XXTRANS, PREDLOSS_RETURN_BETAS);		

		for(j=0; j<p; j++) {
			return_betas_qloss[j] = gsl_matrix_get(QUADLOSS_RETURN_BETAS, j, 0);
			return_betas_ploss[j] = gsl_matrix_get(PREDLOSS_RETURN_BETAS, j, 0);
			for(i=0; i<mcmc_niter; i++) {
				store_betas[i*p+j] = gsl_matrix_get(RETURN_ALL_BETAS,i,j);
				store_lambda_precision_x[i*p+j] = gsl_matrix_get(RETURN_ALL_LAMBDA_PRECISION_X,i,j);
			}
		}
	}

	//More clean-up
	gsl_matrix_free (QUADLOSS_RETURN_BETAS);
	gsl_matrix_free (PREDLOSS_RETURN_BETAS);
	gsl_matrix_free (CURRENT_EXP_XXTRANS);
	gsl_matrix_free (RETURN_EXP_XXTRANS);
	gsl_matrix_free (RETURN_EXP_XXTRANS_BETAS);
	gsl_matrix_free (RETURN_ALL_BETAS);
	gsl_matrix_free (RETURN_ALL_LAMBDA_PRECISION_X);


}

void dagibbs_single_psi_single_nu(int p, int na, int nb,
		const double *ya,const double *yb,const double *xa,const double *wa,const double *wb,
		const double *x_location_factor, const double *x_scale_factor,
		double *imp_xb, double *imp_betas,
		double *imp_beta0, double *imp_sigmasq, 
		double *imp_psi, double *imp_nu, double *imp_tausq, 
		double *imp_mean_x, double *imp_precision_x,
		double *imp_lambda_betas, double *imp_lambda_precision_x,
		int *precision_x_prior_df,
		const int mcmc_burnin, const int mcmc_niter, 
		const int lambda_betas_update, const int lambda_precision_x_update,
		const double lambda_betas_prior_shape, const double lambda_betas_prior_rate,
		double *return_betas_qloss, double *return_betas_ploss,
		double *return_beta0,
		double *store_betas,
		double *store_beta0,
		double *store_sigmasq,
		double *store_tausq,
		double *store_lambda_betas,
		double *store_lambda_precision_x,
		double *store_trace_lambda_precision_x,
		double *store_precision_x_prior_df,
		int *error_code, const int random_seed,
		const int constrain_lambda_precision_x)		
{
	int i, j, k;//index integers

	gsl_set_error_handler_off();

	//Initialize seed
	gsl_rng * ran_num_gen = gsl_rng_alloc(gsl_rng_mt19937);
	gsl_rng_set(ran_num_gen,random_seed);

	//Translate data to GSL matrices
	gsl_matrix * YA = gsl_matrix_calloc(na,1);
	for(i=0; i<na; i++) {gsl_matrix_set(YA, i, 0, ya[i]);}
	gsl_matrix * YB = gsl_matrix_calloc(nb,1);
	for(i=0; i<nb; i++) {gsl_matrix_set(YB, i, 0, yb[i]);}
	gsl_matrix * XA = gsl_matrix_calloc(na,p);
	for(i=0; i<na; i++) for(j=0; j<p; j++) {gsl_matrix_set(XA, i, j, xa[i*p+j]);}
	gsl_matrix * WA = gsl_matrix_calloc(na,p);
	for(i=0; i<na; i++) for(j=0; j<p; j++) {gsl_matrix_set(WA, i, j, wa[i*p+j]);}
	gsl_matrix * WB = gsl_matrix_calloc(nb,p);
	for(i=0; i<nb; i++) for(j=0; j<p; j++) {gsl_matrix_set(WB, i, j, wb[i*p+j]);}
	gsl_matrix * IMP_XB = gsl_matrix_calloc(nb,p);
	for(i=0; i<nb; i++) for(j=0; j<p; j++) {gsl_matrix_set(IMP_XB, i, j, imp_xb[i*p+j]);}
	//

	//Scaled versions
	gsl_matrix * X_LOCATION_FACTOR = gsl_matrix_calloc(p,1);
	for(i=0; i<p; i++) {gsl_matrix_set(X_LOCATION_FACTOR, i, 0, x_location_factor[i]);}
	gsl_matrix * X_INV_SCALE_FACTOR = gsl_matrix_calloc(p,p);
	for(i=0; i<p; i++) {gsl_matrix_set(X_INV_SCALE_FACTOR, i, i, 1./x_scale_factor[i]);}
	gsl_matrix * ONES_LENGTH_NA = gsl_matrix_calloc(na,1);//vector of 1's of length na
	gsl_matrix_set_all(ONES_LENGTH_NA,1.);
	gsl_matrix * ONES_LENGTH_NB = gsl_matrix_calloc(nb,1);//vector of 1's of length nb
	gsl_matrix_set_all(ONES_LENGTH_NB,1.);
	gsl_matrix * S_XA = gsl_matrix_calloc(na,p);
	gsl_matrix_memcpy(S_XA,XA);
	gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,ONES_LENGTH_NA,X_LOCATION_FACTOR,1.,S_XA);
	gsl_blas_dtrmm (CblasRight, CblasLower,CblasNoTrans,CblasNonUnit,1.,X_INV_SCALE_FACTOR,S_XA);
	gsl_matrix * IMP_S_XB = gsl_matrix_calloc(nb,p);
	//

	//Utility values that are used in each iteration
	//xb
	gsl_matrix * MEAN_IMP_XB = gsl_matrix_calloc(nb,p);//mean matrix of imputations
	gsl_matrix * CHOL_PRECISION_IMP_XB = gsl_matrix_calloc(p,p);//Lower triangular L such that Variance(imp_xb[i,]) = (L^-1)^T %*% (L^-1)
	gsl_matrix * INV_CHOL_PRECISION_IMP_XB = gsl_matrix_calloc(p,p);//L^(-1)
	gsl_matrix * RANDOM_VALS_IMP_XB = gsl_matrix_calloc(nb,p);//stochastic component
	gsl_matrix * CROSSPROD_IMP_MEAN_X_IMP_PRECISION_X = gsl_matrix_calloc(1,p);//t(mean_x)%*%precision_x

	//betas
	gsl_matrix * IMP_BETAS = gsl_matrix_calloc(p,1);
	for(i=0; i<p; i++) {gsl_matrix_set(IMP_BETAS, i, 0, imp_betas[i]);}//populate with initial values
	gsl_matrix * SCALED_IMP_BETAS = gsl_matrix_calloc(p,1);//X_SCALE_FACTOR%*%IMP_BETAS
	gsl_matrix * MEAN_SCALED_IMP_BETAS = gsl_matrix_calloc(p,1);//mean matrix (vector) of betas
	gsl_matrix * CHOL_PRECISION_SCALED_IMP_BETAS = gsl_matrix_calloc(p,p);//Lower triangular L such that Variance(imp_betas) = (L^-1)^T %*% (L^-1)
	gsl_matrix * INV_CHOL_PRECISION_SCALED_IMP_BETAS = gsl_matrix_calloc(p,p);//L^(-1)
	gsl_matrix * RANDOM_VALS_SCALED_IMP_BETAS = gsl_matrix_calloc(p,1);//stochastic component
	gsl_matrix * CROSSPROD_S_XA_YA = gsl_matrix_calloc(p,1);//t(s_xa)%*%ya
	gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,S_XA,YA,0.,CROSSPROD_S_XA_YA);
	gsl_matrix * CROSSPROD_S_XA_S_XA = gsl_matrix_calloc(p,p);//t(s_xa)%*%s_xa
	gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,S_XA,S_XA,0.,CROSSPROD_S_XA_S_XA);
	gsl_matrix * CROSSPROD_IMP_S_XB_YB = gsl_matrix_calloc(p,1);//t(s_xb)%*%yb
	gsl_matrix * CROSSPROD_IMP_S_XB_IMP_S_XB = gsl_matrix_calloc(p,p);//t(s_xb)%*%s_xb
	gsl_matrix * COLSUMS_S_XA = gsl_matrix_calloc(p,1);//column vector of column sums of s_xa
	column_sums(na,p,S_XA,COLSUMS_S_XA);
	gsl_matrix * COLSUMS_IMP_S_XB = gsl_matrix_calloc(p,1);//column vector of column sums of imp_s_xb

	//beta0
	gsl_matrix * REG_RESID_A = gsl_matrix_calloc(na,1);//regression residuals: ya - xa%*%betas	
	gsl_matrix * REG_RESID_B = gsl_matrix_calloc(nb,1);//regression residuals: yb - imp_xb%*%betas	
	gsl_matrix * SCALED_OFFSET = gsl_matrix_calloc(p,1);//location_factor^T%*%imp_betas
	gsl_matrix_memcpy(SCALED_OFFSET,IMP_BETAS);
	gsl_matrix_mul_elements(SCALED_OFFSET,X_LOCATION_FACTOR);
	double scaled_imp_beta0 = *imp_beta0 + element_wise_sum(p,1,SCALED_OFFSET,1.);
	double sum_REG_RESID_A = 0.;
	double sum_REG_RESID_B = 0.;

	//sigmasq
	double dotprod_SCALED_IMP_BETAS = 0.;//(SCALED_IMP_BETAS)^T%*%SCALED_IMP_BETAS
	double dotprod_REG_RESID_A = 0.;//(ya-xa%*%imp_betas)^T%*%(ya-xa%*%imp_betas)
	double dotprod_REG_RESID_B= 0.;//(yb-xb%*%imp_betas)^T%*%(yb-xb%*%imp_betas)
	double as_double_n = (double)(na+nb);
	double shape_imp_sigmasq = as_double_n / 2.;//shape parameter for posterior distribution
	if(*imp_lambda_betas > 0.) {shape_imp_sigmasq += (double)p / 2.;}
	double scale_imp_sigmasq;

	//psi
	double sum_XA = element_wise_sum(na,p,XA,1.);
	double sum_WA = element_wise_sum(na,p,WA,1.);
	double sum_WB = element_wise_sum(nb,p,WB,1.);
	double sum_IMP_XB = 0.;
	double as_double_n_times_p = (double)((na+nb)*p);

	//nu
	gsl_matrix * CROSSPROD_XA_XA = gsl_matrix_calloc(p,p);//t(xa)%*%xa
	gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,XA,XA,0.,CROSSPROD_XA_XA);
	gsl_matrix * CROSSPROD_XA_WA = gsl_matrix_calloc(p,p);//(xa)^T%*%wa
	gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,XA,WA,0.,CROSSPROD_XA_WA);
	double trace_CROSSPROD_XA_XA = matrix_trace(p,CROSSPROD_XA_XA);//trace((xa)^T%*%xa)
	double trace_CROSSPROD_XA_WA = matrix_trace(p,CROSSPROD_XA_WA);//trace((xa)^T%*%wa)
	gsl_matrix * CROSSPROD_IMP_XB_IMP_XB = gsl_matrix_calloc(p,p);//t(xb)%*%xb
	gsl_matrix * CROSSPROD_IMP_XB_WB = gsl_matrix_calloc(p,p);//(xb)^T%*%wb
	double numerator_mean_imp_nu;
	double denominator_mean_imp_nu;
	double sd_imp_nu;
	
	//tausq
	gsl_matrix * ME_RESID_A = gsl_matrix_calloc(na,p);//measurement error residuals: wa-imp_nu*xa
	gsl_matrix * ME_RESID_B = gsl_matrix_calloc(nb,p);//measurement error residuals: wb-imp_nu*imp_xb
	double shape_imp_tausq = as_double_n_times_p / 2.;
	double scale_imp_tausq;

	//mean_x
	gsl_matrix * IMP_MEAN_X = gsl_matrix_calloc(p,1);
	for(i=0; i<p; i++) {gsl_matrix_set(IMP_MEAN_X, i, 0, imp_mean_x[i]);}//populate with initial values
	gsl_matrix * RANDOM_VALS_IMP_MEAN_X = gsl_matrix_calloc(p,1);//stochastic component
	gsl_matrix * CHOL_IMP_PRECISION_X = gsl_matrix_calloc(p,p);
	gsl_matrix * INV_CHOL_IMP_PRECISION_X = gsl_matrix_calloc(p,p);

	//precision_x
	gsl_matrix * COLSUMS_XA = gsl_matrix_calloc(p,1);//column vector of column sums of xa
	column_sums(na,p,XA,COLSUMS_XA);
	gsl_matrix * COLSUMS_IMP_XB = gsl_matrix_calloc(p,1);//column vector of column sums of imp_xb
	gsl_matrix * IMP_PRECISION_X = gsl_matrix_calloc(p,p);
	for(i=0; i<p; i++) for(j=0; j<p; j++) {gsl_matrix_set(IMP_PRECISION_X, i, j, imp_precision_x[i*p+j]);}//populate with initial values
	int df_IMP_PRECISION_X;
	gsl_matrix * CHOL_INVSCALE_IMP_PRECISION_X = gsl_matrix_calloc(p,p);
	gsl_matrix * INV_CHOL_INVSCALE_IMP_PRECISION_X = gsl_matrix_calloc(p,p);
	double lndet_IMP_PRECISION_X;

	//lambda_betas
	int n_ebiter_betas = 10;
	int m1 = 0;
	double store_dotprod_SCALED_IMP_BETAS_over_imp_sigmasq = 0.;

	//lambda_precision_x
	gsl_matrix * IMP_LAMBDA_PRECISION_X = gsl_matrix_calloc(p,p);
	for(i=0; i<p; i++) for(j=0; j<p; j++) {gsl_matrix_set(IMP_LAMBDA_PRECISION_X, i, j, imp_lambda_precision_x[i*p+j]);}//populate with initial values
	int n_ebiter_precision_X = (int)(ceil(p/4));
	int m2 = 0;
	gsl_matrix * STORE_PRECISION_X = gsl_matrix_calloc(p,p);
	double store_log_det_PRECISION_X = 0.;	
	double trace_IMP_LAMBDA_PRECISION_X = matrix_trace(p, IMP_LAMBDA_PRECISION_X);
	int prior_df_grid[10*p];
	for(i=0;i<10*p;i++) {
		prior_df_grid[i] = *precision_x_prior_df-(5*p)+(i+1);	
	}

	//Return values
	*return_beta0 = 0.;
	gsl_matrix * QUADLOSS_RETURN_BETAS = gsl_matrix_calloc(p,1);
	gsl_matrix * PREDLOSS_RETURN_BETAS = gsl_matrix_calloc(p,1);
	gsl_matrix * CURRENT_EXP_XXTRANS = gsl_matrix_calloc(p,p);
	gsl_matrix * RETURN_EXP_XXTRANS = gsl_matrix_calloc(p,p);
	gsl_matrix * RETURN_EXP_XXTRANS_BETAS = gsl_matrix_calloc(p,1);
	gsl_matrix * RETURN_ALL_BETAS = gsl_matrix_calloc(mcmc_niter,p);
	gsl_matrix * RETURN_ALL_LAMBDA_PRECISION_X = gsl_matrix_calloc(mcmc_niter,p);

	//
	
	for(k = 0; k<(mcmc_burnin+mcmc_niter);k++) {

		//////////////////////
		//Imputation of xb//
		/////////////////////

		//First draw stochastic component
		for (i=0; i<nb; i++) for(j=0; j<p; j++) {gsl_matrix_set (RANDOM_VALS_IMP_XB, i, j, gsl_ran_ugaussian(ran_num_gen));}
			
		//Now determine the covariance matrix (via the inverse of the cholesky decomposition of the precision)
		// precision <- identity
		gsl_matrix_set_identity(CHOL_PRECISION_IMP_XB);
		gsl_matrix_scale(CHOL_PRECISION_IMP_XB,pow(*imp_nu,2.));
		// precision <- beta*beta^T/sigma2 + nu^2/tausq*precision
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,(1./ *imp_sigmasq),IMP_BETAS,IMP_BETAS,(1. / *imp_tausq),CHOL_PRECISION_IMP_XB);
		// precision <- precision + precision_x
		gsl_matrix_add(CHOL_PRECISION_IMP_XB,IMP_PRECISION_X);
		//take sqrt, will use only lower triangular values (L)
		*error_code = gsl_linalg_cholesky_decomp(CHOL_PRECISION_IMP_XB);
		if(*error_code == GSL_EDOM) {*error_code=1; break;}
		//now invert (L^(-1)
		gsl_matrix_set_identity(INV_CHOL_PRECISION_IMP_XB);
		gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1., CHOL_PRECISION_IMP_XB, INV_CHOL_PRECISION_IMP_XB);


		//Now determine the mean of xb
		//mean <- wb
		gsl_matrix_memcpy(MEAN_IMP_XB,WB);
		//mean <- mean - psi
		gsl_matrix_add_constant(MEAN_IMP_XB,-1. * (*imp_psi));
		//mean <- yb*beta^T/sigma2 + nu/tausq*mean
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,(1./ *imp_sigmasq),YB,IMP_BETAS,(*imp_nu/ *imp_tausq),MEAN_IMP_XB);
		//mean <- mean - beta0/sigma2*beta
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-(*imp_beta0 / *imp_sigmasq),ONES_LENGTH_NB,IMP_BETAS,1.,MEAN_IMP_XB);
		//mean <- mean + 1*mu_x^T*precision_x
		gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,IMP_MEAN_X,IMP_PRECISION_X,0.,CROSSPROD_IMP_MEAN_X_IMP_PRECISION_X);
		gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,1,ONES_LENGTH_NB,CROSSPROD_IMP_MEAN_X_IMP_PRECISION_X,1.,MEAN_IMP_XB);
		//mean <- mean*(L^-1)^T
		gsl_blas_dtrmm (CblasRight, CblasLower,CblasTrans,CblasNonUnit,1.,INV_CHOL_PRECISION_IMP_XB,MEAN_IMP_XB);
		//mean <- mean*(L^-1)
		gsl_blas_dtrmm (CblasRight, CblasLower,CblasNoTrans,CblasNonUnit,1.,INV_CHOL_PRECISION_IMP_XB,MEAN_IMP_XB);

		// put xb <- mean + random_vals * sqrt(variance))
		gsl_matrix_memcpy(IMP_XB,MEAN_IMP_XB);
		gsl_blas_dtrmm(CblasRight,CblasLower,CblasNoTrans,CblasNonUnit,1.,INV_CHOL_PRECISION_IMP_XB,RANDOM_VALS_IMP_XB);
		gsl_matrix_add(IMP_XB,RANDOM_VALS_IMP_XB);

		//Store colsums
		column_sums(nb,p,IMP_XB,COLSUMS_IMP_XB);

		//calculate IMP_S_XB
		gsl_matrix_memcpy(IMP_S_XB,IMP_XB);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,ONES_LENGTH_NB,X_LOCATION_FACTOR,1.,IMP_S_XB);
		gsl_blas_dtrmm (CblasRight, CblasLower,CblasNoTrans,CblasNonUnit,1.,X_INV_SCALE_FACTOR,IMP_S_XB);

		//////////////////////
		//Imputation of scaled_betas//
		/////////////////////

		column_sums(nb,p,IMP_S_XB,COLSUMS_IMP_S_XB);
		gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,IMP_S_XB,IMP_S_XB,0.,CROSSPROD_IMP_S_XB_IMP_S_XB);
		gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,IMP_S_XB,YB,0.,CROSSPROD_IMP_S_XB_YB);

		//First draw stochastic component
		for(i=0; i<p; i++) {gsl_matrix_set (RANDOM_VALS_SCALED_IMP_BETAS, i, 0, gsl_ran_ugaussian(ran_num_gen));}
		
		//Now the covariance matrix (via the precision)
		// put precision <- imp_lambda_betas*Ip+(s_xa)^T%*%(s_xa)+(imp_s_xb)^T%*%imp_s_xb
		gsl_matrix_set_identity(CHOL_PRECISION_SCALED_IMP_BETAS);
		gsl_matrix_scale(CHOL_PRECISION_SCALED_IMP_BETAS,*imp_lambda_betas);
		gsl_matrix_add(CHOL_PRECISION_SCALED_IMP_BETAS,CROSSPROD_S_XA_S_XA);	
		gsl_matrix_add(CHOL_PRECISION_SCALED_IMP_BETAS,CROSSPROD_IMP_S_XB_IMP_S_XB);


		//take sqrt, will use only lower triangular values (L)
		*error_code = gsl_linalg_cholesky_decomp(CHOL_PRECISION_SCALED_IMP_BETAS);
		if(*error_code == GSL_EDOM) {*error_code=2; break;}
		//now invert (L^(-1)
		gsl_matrix_set_identity(INV_CHOL_PRECISION_SCALED_IMP_BETAS);
		gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1., CHOL_PRECISION_SCALED_IMP_BETAS, INV_CHOL_PRECISION_SCALED_IMP_BETAS);

		//Now the mean
		//mean <- -beta0*(colsums(s_xa)+colsums(imp_s_xb)) + (s_xa)^T%*%ya + (imp_s_xb)^T%*%yb 
		gsl_matrix_memcpy(MEAN_SCALED_IMP_BETAS,COLSUMS_S_XA);
		gsl_matrix_add(MEAN_SCALED_IMP_BETAS,COLSUMS_IMP_S_XB);
		gsl_matrix_scale(MEAN_SCALED_IMP_BETAS,-(scaled_imp_beta0));
		gsl_matrix_add(MEAN_SCALED_IMP_BETAS,CROSSPROD_S_XA_YA);
		gsl_matrix_add(MEAN_SCALED_IMP_BETAS,CROSSPROD_IMP_S_XB_YB);
		//mean <- (L^-1)*mean
		gsl_blas_dtrmm (CblasLeft, CblasLower,CblasNoTrans,CblasNonUnit,1.,INV_CHOL_PRECISION_SCALED_IMP_BETAS,MEAN_SCALED_IMP_BETAS);
		//mean <- (L^-1)^T*mean
		gsl_blas_dtrmm (CblasLeft, CblasLower,CblasTrans,CblasNonUnit,1.,INV_CHOL_PRECISION_SCALED_IMP_BETAS,MEAN_SCALED_IMP_BETAS);

		// put imp_betas <- mean + sqrt(sigmasq)*sqrt(variance)*random_vals )
		gsl_matrix_memcpy(SCALED_IMP_BETAS,MEAN_SCALED_IMP_BETAS);
		gsl_blas_dtrmm(CblasLeft,CblasLower,CblasTrans,CblasNonUnit,sqrt(*imp_sigmasq),INV_CHOL_PRECISION_SCALED_IMP_BETAS,RANDOM_VALS_SCALED_IMP_BETAS);
		gsl_matrix_add(SCALED_IMP_BETAS,RANDOM_VALS_SCALED_IMP_BETAS);

		//imp_betas (original scale)
		gsl_matrix_memcpy(IMP_BETAS,SCALED_IMP_BETAS);
		gsl_blas_dtrmm (CblasLeft, CblasLower,CblasNoTrans,CblasNonUnit,1.,X_INV_SCALE_FACTOR,IMP_BETAS);
		
		//////////////////////
		//Imputation of beta0
		/////////////////////

		gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,-1.,XA,IMP_BETAS,0.,REG_RESID_A);
		gsl_matrix_add (REG_RESID_A, YA);
		sum_REG_RESID_A = element_wise_sum(na,1,REG_RESID_A,1.);

		gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,-1.,IMP_XB,IMP_BETAS,0.,REG_RESID_B);
		gsl_matrix_add (REG_RESID_B, YB);
		sum_REG_RESID_B = element_wise_sum(nb,1,REG_RESID_B,1.);

		*imp_beta0 = (sum_REG_RESID_A+sum_REG_RESID_B)/as_double_n + sqrt(*imp_sigmasq/as_double_n)*gsl_ran_ugaussian(ran_num_gen);
		gsl_matrix_memcpy(SCALED_OFFSET,IMP_BETAS);
		gsl_matrix_mul_elements(SCALED_OFFSET,X_LOCATION_FACTOR);
		scaled_imp_beta0 = *imp_beta0 + element_wise_sum(p,1,SCALED_OFFSET,1.);

		//////////////////////
		//Imputation of sigmasq//
		/////////////////////

		dotprod_SCALED_IMP_BETAS = element_wise_sum(p,1,SCALED_IMP_BETAS,2.);

		gsl_matrix_add_constant (REG_RESID_A, -(*imp_beta0));
		dotprod_REG_RESID_A = element_wise_sum(na,1,REG_RESID_A,2.);

		gsl_matrix_add_constant (REG_RESID_B, -(*imp_beta0));
		dotprod_REG_RESID_B= element_wise_sum(nb,1,REG_RESID_B,2.);

		scale_imp_sigmasq = 2 / ((*imp_lambda_betas) * (dotprod_SCALED_IMP_BETAS) + dotprod_REG_RESID_A + dotprod_REG_RESID_B);
		*imp_sigmasq = 1./gsl_ran_gamma(ran_num_gen,shape_imp_sigmasq,scale_imp_sigmasq);
	
		//////////////////////
		//Imputation of psi//
		/////////////////////

		sum_IMP_XB = element_wise_sum(nb,p,IMP_XB,1.);
		
		*imp_psi = (sum_WA + sum_WB - (*imp_nu)*(sum_XA + sum_IMP_XB))/as_double_n_times_p + sqrt((*imp_tausq))/sqrt(as_double_n_times_p) * gsl_ran_ugaussian(ran_num_gen);

		//////////////////////
		//Imputation of nu//
		/////////////////////
		
		gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,IMP_XB,IMP_XB,0.,CROSSPROD_IMP_XB_IMP_XB);
		gsl_blas_dgemm(CblasTrans,CblasNoTrans,1.,IMP_XB,WB,0.,CROSSPROD_IMP_XB_WB);		
		numerator_mean_imp_nu = trace_CROSSPROD_XA_WA + matrix_trace(p,CROSSPROD_IMP_XB_WB) - (*imp_psi) * (sum_XA + sum_IMP_XB);
		denominator_mean_imp_nu = trace_CROSSPROD_XA_XA + matrix_trace(p,CROSSPROD_IMP_XB_IMP_XB);
		sd_imp_nu = sqrt((*imp_tausq))/sqrt(denominator_mean_imp_nu);
		*imp_nu =  (numerator_mean_imp_nu/denominator_mean_imp_nu) + sd_imp_nu * gsl_ran_ugaussian(ran_num_gen);
	
		//////////////////////
		//Imputation of tausq//
		/////////////////////

		gsl_matrix_memcpy(ME_RESID_A,XA);
		gsl_matrix_scale (ME_RESID_A, -1. * (*imp_nu));
		gsl_matrix_add_constant(ME_RESID_A,-1. * (*imp_psi));
		gsl_matrix_add(ME_RESID_A,WA);

		gsl_matrix_memcpy(ME_RESID_B,IMP_XB);
		gsl_matrix_scale (ME_RESID_B, -1. * (*imp_nu));
		gsl_matrix_add_constant(ME_RESID_B,-1. * (*imp_psi));		
		gsl_matrix_add(ME_RESID_B,WB);
		scale_imp_tausq = 2. / (element_wise_sum(na,p,ME_RESID_A,2.)+element_wise_sum(nb,p,ME_RESID_B,2.));
		*imp_tausq = 1./gsl_ran_gamma(ran_num_gen,shape_imp_tausq,scale_imp_tausq);
	
		//////////////////////
		//Imputation of precision_x//
		/////////////////////
		
		df_IMP_PRECISION_X = (int)(na + nb);
		if(trace_IMP_LAMBDA_PRECISION_X>0.) {df_IMP_PRECISION_X += *precision_x_prior_df;}

		// put CHOL_INVSCALE_IMP_PRECISION_X <- (xa - 1_na %*% mean_x^T)^T %*% (xa - 1_na %*% mean_x^T)
		// by adding each term in the cross product			
		gsl_matrix_memcpy(CHOL_INVSCALE_IMP_PRECISION_X,CROSSPROD_XA_XA);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,IMP_MEAN_X,COLSUMS_XA,1.,CHOL_INVSCALE_IMP_PRECISION_X);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,COLSUMS_XA,IMP_MEAN_X,1.,CHOL_INVSCALE_IMP_PRECISION_X);

		// put CHOL_INVSCALE_IMP_PRECISION_X <- CHOL_INVSCALE_IMP_PRECISION_X + (imp_xb - 1_nb %*% mean_x^T)^T %*% (imp_xb - 1_nb %*% mean_x^T)
		//by adding each term in the cross product
		gsl_matrix_add(CHOL_INVSCALE_IMP_PRECISION_X,CROSSPROD_IMP_XB_IMP_XB);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,IMP_MEAN_X,COLSUMS_IMP_XB,1.,CHOL_INVSCALE_IMP_PRECISION_X);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,-1.,COLSUMS_IMP_XB,IMP_MEAN_X,1.,CHOL_INVSCALE_IMP_PRECISION_X);
		gsl_blas_dgemm(CblasNoTrans,CblasTrans,as_double_n,IMP_MEAN_X,IMP_MEAN_X,1.,CHOL_INVSCALE_IMP_PRECISION_X);

		// put CHOL_INVSCALE_IMP_PRECISION_X <- CHOL_INVSCALE_IMP_PRECISION_X + IMP_LAMBDA_PRECISION_X
		gsl_matrix_add(CHOL_INVSCALE_IMP_PRECISION_X,IMP_LAMBDA_PRECISION_X);

		//take sqrt, will use only lower triangular values (L)
		*error_code = gsl_linalg_cholesky_decomp(CHOL_INVSCALE_IMP_PRECISION_X);
		if(*error_code == GSL_EDOM) {*error_code=3; break;}
		//now invert (L^(-1)
		gsl_matrix_set_identity(INV_CHOL_INVSCALE_IMP_PRECISION_X);
		gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1., CHOL_INVSCALE_IMP_PRECISION_X, INV_CHOL_INVSCALE_IMP_PRECISION_X);
		
		//Pass this to wishart to draw a random wishart
		lndet_IMP_PRECISION_X = wishart(df_IMP_PRECISION_X, INV_CHOL_INVSCALE_IMP_PRECISION_X, p, ran_num_gen, IMP_PRECISION_X);

		//////////////////////
		//Imputation of mean_x//
		/////////////////////

		//First draw stochastic component
		for(i=0; i<p; i++) {gsl_matrix_set (RANDOM_VALS_IMP_MEAN_X, i, 0, gsl_ran_ugaussian(ran_num_gen));}
		
		//take sqrt of precision_x, will use only lower triangular values (L)
		gsl_matrix_memcpy(CHOL_IMP_PRECISION_X,IMP_PRECISION_X);
		*error_code = gsl_linalg_cholesky_decomp(CHOL_IMP_PRECISION_X);
		if(*error_code == GSL_EDOM) {*error_code=4; break;}		
		//now invert (L^-1)
		gsl_matrix_set_identity(INV_CHOL_IMP_PRECISION_X);
		gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1., CHOL_IMP_PRECISION_X, INV_CHOL_IMP_PRECISION_X);	
	
		//put imp_mean_x <- (colsums(xa)+colsums(imp_xb))/(na+nb)
		gsl_matrix_memcpy(IMP_MEAN_X,COLSUMS_XA);
		gsl_matrix_add(IMP_MEAN_X,COLSUMS_IMP_XB);
		gsl_matrix_scale(IMP_MEAN_X,1./as_double_n);

		//Now add the stochastic component (ie put imp_mean_x <- imp_mean_x + sqrt(variance)*random_seed )
		gsl_blas_dtrmm(CblasLeft,CblasLower,CblasTrans,CblasNonUnit,1./sqrt(as_double_n),INV_CHOL_IMP_PRECISION_X,RANDOM_VALS_IMP_MEAN_X);
		gsl_matrix_add(IMP_MEAN_X,RANDOM_VALS_IMP_MEAN_X);


		//////////////////////
		//Imputation of lambda_betas//
		/////////////////////
 		if(lambda_betas_update==1) {//empirical bayes update
			m1 += 1;
			store_dotprod_SCALED_IMP_BETAS_over_imp_sigmasq += dotprod_SCALED_IMP_BETAS/(*imp_sigmasq);
			if(m1==n_ebiter_betas) {
				if(store_dotprod_SCALED_IMP_BETAS_over_imp_sigmasq==0.) {
					*error_code=5; 
					break;
				} else {
					*imp_lambda_betas = (double)((n_ebiter_betas * p)/store_dotprod_SCALED_IMP_BETAS_over_imp_sigmasq);
				}
				store_dotprod_SCALED_IMP_BETAS_over_imp_sigmasq = 0.;
				m1 = 0;
				if(k>=mcmc_burnin) {n_ebiter_betas=50;}
			}
		} else if(lambda_betas_update==2) {//hierarchical update
			*imp_lambda_betas = gsl_ran_gamma(ran_num_gen,(double)(p/2)+lambda_betas_prior_shape,1./((dotprod_SCALED_IMP_BETAS/(*imp_sigmasq))/2.+lambda_betas_prior_rate));
		}

		//////////////////////
		//Imputation of lambda_precision_x//
		/////////////////////
		if(lambda_precision_x_update==1) {
			m2 += 1;
			if(constrain_lambda_precision_x==1) {
				gsl_matrix_add_diagonals(p,STORE_PRECISION_X,IMP_PRECISION_X);
			} else if(constrain_lambda_precision_x==2) {
				gsl_matrix_add_constant_diagonal_only(p,STORE_PRECISION_X,matrix_trace(p, IMP_PRECISION_X)/(double)(p));
			}
			store_log_det_PRECISION_X += lndet_IMP_PRECISION_X;
			if(m2==n_ebiter_precision_X) {
				gsl_matrix_set_identity(IMP_LAMBDA_PRECISION_X);
				*error_code = gsl_linalg_cholesky_decomp(STORE_PRECISION_X);
				if(*error_code == GSL_EDOM) {*error_code=6; break;}
				gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1., STORE_PRECISION_X, IMP_LAMBDA_PRECISION_X);	
				gsl_blas_dtrsm(CblasLeft, CblasLower, CblasTrans, CblasNonUnit, (double)(*precision_x_prior_df*n_ebiter_precision_X), STORE_PRECISION_X, IMP_LAMBDA_PRECISION_X);
				gsl_matrix_set_zero(STORE_PRECISION_X);
				trace_IMP_LAMBDA_PRECISION_X = matrix_trace(p, IMP_LAMBDA_PRECISION_X);
				m2 = 0;
				if(k>=1000) {n_ebiter_precision_X=3*p;}
			}
		} else if(lambda_precision_x_update==2) {
			m2 += 1;
			store_log_det_PRECISION_X += lndet_IMP_PRECISION_X;
			if(m2==n_ebiter_precision_X) {
				*precision_x_prior_df = best_df(prior_df_grid,10*p,p*log(2)+gsl_linalg_LU_lndet(IMP_LAMBDA_PRECISION_X)+(store_log_det_PRECISION_X/n_ebiter_precision_X),p);
				for(i=0;i<10*p;i++) {
					prior_df_grid[i] = *precision_x_prior_df-(5*p)+(i+1);
				}
				store_log_det_PRECISION_X = 0.;
				m2 = 0;
				if(k>=1000) {n_ebiter_precision_X=3*p;}
			}
		}
	

		if(k>=mcmc_burnin) {
			gsl_matrix_add(QUADLOSS_RETURN_BETAS,IMP_BETAS);
			gsl_matrix_set_identity(CURRENT_EXP_XXTRANS);
			gsl_blas_dtrmm(CblasRight, CblasLower, CblasTrans, CblasNonUnit, 1., INV_CHOL_IMP_PRECISION_X, CURRENT_EXP_XXTRANS);		
			gsl_blas_dtrmm(CblasRight, CblasLower, CblasNoTrans, CblasNonUnit, 1., INV_CHOL_IMP_PRECISION_X, CURRENT_EXP_XXTRANS);		
			gsl_blas_dgemm(CblasNoTrans,CblasTrans,1.,IMP_MEAN_X,IMP_MEAN_X,1.,CURRENT_EXP_XXTRANS);
			gsl_matrix_add(RETURN_EXP_XXTRANS,CURRENT_EXP_XXTRANS);
			gsl_blas_dgemm(CblasNoTrans,CblasNoTrans,1.,CURRENT_EXP_XXTRANS,IMP_BETAS,1.,RETURN_EXP_XXTRANS_BETAS);	
			
			*return_beta0 += (*imp_beta0) / (double)(mcmc_niter);

			gsl_vector_view VIEW_AS_VECTOR_IMP_BETAS = gsl_matrix_column (IMP_BETAS, 0);
			gsl_matrix_set_row(RETURN_ALL_BETAS,k-mcmc_burnin, &VIEW_AS_VECTOR_IMP_BETAS.vector);
			store_beta0[k-mcmc_burnin] = (*imp_beta0);
			store_sigmasq[k-mcmc_burnin] = (*imp_sigmasq);
			store_tausq[k-mcmc_burnin] = (*imp_tausq);
			store_lambda_betas[k-mcmc_burnin] = (*imp_lambda_betas);
			store_trace_lambda_precision_x[k-mcmc_burnin] = trace_IMP_LAMBDA_PRECISION_X;
			store_precision_x_prior_df[k-mcmc_burnin] = *precision_x_prior_df;
			gsl_vector_view VIEW_AS_VECTOR_DIAGONAL_IMP_LAMBDA_PRECISION_X = gsl_matrix_diagonal (IMP_LAMBDA_PRECISION_X);
			gsl_matrix_set_row(RETURN_ALL_LAMBDA_PRECISION_X,k-mcmc_burnin, &VIEW_AS_VECTOR_DIAGONAL_IMP_LAMBDA_PRECISION_X.vector);
		}

	}

	////////////////
	//Clean up
	///////////////
	

	gsl_rng_free (ran_num_gen);

	//data
	gsl_matrix_free (YA );
	gsl_matrix_free (YB );
	gsl_matrix_free (XA );		
	gsl_matrix_free (WA );
	gsl_matrix_free (WB );
	gsl_matrix_free (IMP_XB );

	//scaled versions
	gsl_matrix_free (X_INV_SCALE_FACTOR);
	gsl_matrix_free (X_LOCATION_FACTOR);
	gsl_matrix_free (ONES_LENGTH_NA);
	gsl_matrix_free (ONES_LENGTH_NB);
	gsl_matrix_free (S_XA );
	gsl_matrix_free (IMP_S_XB );

	//xb
	gsl_matrix_free (MEAN_IMP_XB );
	gsl_matrix_free (CHOL_PRECISION_IMP_XB );
	gsl_matrix_free (INV_CHOL_PRECISION_IMP_XB );
	gsl_matrix_free (RANDOM_VALS_IMP_XB );
	gsl_matrix_free (CROSSPROD_IMP_MEAN_X_IMP_PRECISION_X );

	//betas
	gsl_matrix_free (IMP_BETAS );
	gsl_matrix_free (SCALED_IMP_BETAS );
	gsl_matrix_free (MEAN_SCALED_IMP_BETAS );
	gsl_matrix_free (CHOL_PRECISION_SCALED_IMP_BETAS );
	gsl_matrix_free (INV_CHOL_PRECISION_SCALED_IMP_BETAS );
	gsl_matrix_free (RANDOM_VALS_SCALED_IMP_BETAS );
	gsl_matrix_free (CROSSPROD_S_XA_YA);
	gsl_matrix_free (CROSSPROD_S_XA_S_XA);
	gsl_matrix_free (CROSSPROD_IMP_S_XB_YB );
	gsl_matrix_free (CROSSPROD_IMP_S_XB_IMP_S_XB );
	gsl_matrix_free (COLSUMS_IMP_S_XB );
	gsl_matrix_free (COLSUMS_S_XA);

	//beta0
	gsl_matrix_free (REG_RESID_A );
	gsl_matrix_free (REG_RESID_B );
	gsl_matrix_free (SCALED_OFFSET );

	//psi

	//nu
	gsl_matrix_free (CROSSPROD_XA_XA);
	gsl_matrix_free (CROSSPROD_XA_WA);
	gsl_matrix_free (CROSSPROD_IMP_XB_IMP_XB);
	gsl_matrix_free (CROSSPROD_IMP_XB_WB);

	//tau2
	gsl_matrix_free (ME_RESID_A );
	gsl_matrix_free (ME_RESID_B );

	//mean_x
	gsl_matrix_free (IMP_MEAN_X);
	gsl_matrix_free (RANDOM_VALS_IMP_MEAN_X );
	gsl_matrix_free (CHOL_IMP_PRECISION_X );
	gsl_matrix_free (INV_CHOL_IMP_PRECISION_X );

	//precision_x
	gsl_matrix_free (COLSUMS_XA);
	gsl_matrix_free (COLSUMS_IMP_XB);
	gsl_matrix_free (IMP_PRECISION_X);
	gsl_matrix_free (CHOL_INVSCALE_IMP_PRECISION_X  );
	gsl_matrix_free (INV_CHOL_INVSCALE_IMP_PRECISION_X );

	//lambda_precision_x
	gsl_matrix_free (IMP_LAMBDA_PRECISION_X);
	gsl_matrix_free (STORE_PRECISION_X);


	if(*error_code==0) {

		gsl_matrix_scale(QUADLOSS_RETURN_BETAS,1./(double)(mcmc_niter));

		gsl_matrix_memcpy(PREDLOSS_RETURN_BETAS,RETURN_EXP_XXTRANS_BETAS);
		gsl_linalg_cholesky_decomp(RETURN_EXP_XXTRANS);
		gsl_blas_dtrsm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1., RETURN_EXP_XXTRANS, PREDLOSS_RETURN_BETAS);	
		gsl_blas_dtrsm(CblasLeft, CblasLower, CblasTrans, CblasNonUnit, 1., RETURN_EXP_XXTRANS, PREDLOSS_RETURN_BETAS);		

		for(j=0; j<p; j++) {
			return_betas_qloss[j] = gsl_matrix_get(QUADLOSS_RETURN_BETAS, j, 0);
			return_betas_ploss[j] = gsl_matrix_get(PREDLOSS_RETURN_BETAS, j, 0);
			for(i=0; i<mcmc_niter; i++) {
				store_betas[i*p+j] = gsl_matrix_get(RETURN_ALL_BETAS,i,j);
				store_lambda_precision_x[i*p+j] = gsl_matrix_get(RETURN_ALL_LAMBDA_PRECISION_X,i,j);
			}
		}
	}

	//More clean-up
	gsl_matrix_free (QUADLOSS_RETURN_BETAS);
	gsl_matrix_free (PREDLOSS_RETURN_BETAS);
	gsl_matrix_free (CURRENT_EXP_XXTRANS);
	gsl_matrix_free (RETURN_EXP_XXTRANS);
	gsl_matrix_free (RETURN_EXP_XXTRANS_BETAS);
	gsl_matrix_free (RETURN_ALL_BETAS);
	gsl_matrix_free (RETURN_ALL_LAMBDA_PRECISION_X);

}

void R_dagibbs(int *p, int *na, int *nb,
		double *ya,double *yb, double *xa,double *wa, double *wb,
		double *x_location_factor, double *x_scale_factor,
		double *imp_xb, double *imp_betas,
		double *imp_beta0, double *imp_sigmasq, 
		double *imp_psi, double *imp_nu, double *imp_tausq,
		double *imp_mean_x, double *imp_precision_x,
		double *imp_lambda_betas, double *imp_lambda_precision_x,
		int *precision_x_prior_df,
		int *mcmc_burnin, int *mcmc_niter, 
		int *lambda_betas_update, int *lambda_precision_x_update,
		double *lambda_betas_prior_shape, double *lambda_betas_prior_rate,
		double *return_betas_qloss, double *return_betas_ploss,
		double *return_beta0,
		double *store_betas,
		double *store_beta0,
		double *store_sigmasq,
		double *store_tausq,
		double *store_lambda_betas,
		double *store_lambda_precision_x,
		double *store_trace_lambda_precision_x,
		double *store_precision_x_prior_df,
		int *error_code,int *random_seed,
		int *single_psi, int *single_nu,
		int *constrain_lambda_precision_x)
{
	if((*single_psi)+(*single_nu)<=1) {
		dagibbs(*p,*na,*nb,
			ya,yb,xa,wa,wb,
			x_location_factor,x_scale_factor,
			imp_xb,imp_betas,
			imp_beta0,imp_sigmasq,
			imp_psi,imp_nu,imp_tausq,
			imp_mean_x,imp_precision_x,
			imp_lambda_betas,imp_lambda_precision_x,
			precision_x_prior_df,
			*mcmc_burnin, *mcmc_niter,
			*lambda_betas_update,*lambda_precision_x_update,
			*lambda_betas_prior_shape,*lambda_betas_prior_rate,
			return_betas_qloss,return_betas_ploss,
			return_beta0,
			store_betas,
			store_beta0,
			store_sigmasq,
			store_tausq,
			store_lambda_betas,
			store_lambda_precision_x,
			store_trace_lambda_precision_x,
			store_precision_x_prior_df,
			error_code,*random_seed,
			*constrain_lambda_precision_x);
	} else {
		dagibbs_single_psi_single_nu(*p,*na,*nb,
			ya,yb,xa,wa,wb,
			x_location_factor,x_scale_factor,
			imp_xb,imp_betas,
			imp_beta0,imp_sigmasq,
			imp_psi,imp_nu,imp_tausq,
			imp_mean_x,imp_precision_x,
			imp_lambda_betas,imp_lambda_precision_x,
			precision_x_prior_df,
			*mcmc_burnin,*mcmc_niter,
			*lambda_betas_update,*lambda_precision_x_update,
			*lambda_betas_prior_shape,*lambda_betas_prior_rate,
			return_betas_qloss,return_betas_ploss,
			return_beta0,
			store_betas,
			store_beta0,
			store_sigmasq,
			store_tausq,
			store_lambda_betas,
			store_lambda_precision_x,
			store_trace_lambda_precision_x,
			store_precision_x_prior_df,
			error_code,*random_seed,
			*constrain_lambda_precision_x);
	}
}


